import pytest

from blends.models import (
    SyntaxGraph,
)
from blends.stack.attributes import (
    STACK_GRAPH_IS_DEFINITION,
    STACK_GRAPH_IS_EXPORTED,
    STACK_GRAPH_IS_REFERENCE,
    STACK_GRAPH_KIND,
    STACK_GRAPH_SCOPE,
    STACK_GRAPH_SYMBOL,
)
from blends.stack.criteria import (
    is_complete_definition_path,
    is_definition_endpoint,
    is_endpoint,
    is_exported_scope_endpoint,
    is_jump_to_boundary,
    is_reference_endpoint,
    is_root_endpoint,
)
from blends.stack.node_kinds import (
    StackGraphNodeKind,
)
from blends.stack.stacks import (
    ScopeStackNode,
    StackState,
    SymbolStackNode,
)
from blends.stack.transitions import (
    apply_node,
)
from test.stack.test_helpers import view_for_graph


@pytest.mark.blends_test_group("stack_unittesting")
def test_is_reference_endpoint_push_with_flag_true() -> None:
    graph = SyntaxGraph()
    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SYMBOL.value)
    graph.update_node_attribute("2", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("2", STACK_GRAPH_IS_REFERENCE, value=True)

    view = view_for_graph(graph)
    node_index = view.nid_to_index["2"]

    assert is_reference_endpoint(view, node_index)


@pytest.mark.blends_test_group("stack_unittesting")
def test_is_reference_endpoint_push_with_flag_false() -> None:
    graph = SyntaxGraph()
    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SYMBOL.value)
    graph.update_node_attribute("2", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("2", STACK_GRAPH_IS_REFERENCE, value=False)

    view = view_for_graph(graph)
    node_index = view.nid_to_index["2"]

    assert not is_reference_endpoint(view, node_index)
    assert not is_endpoint(view, node_index)


@pytest.mark.blends_test_group("stack_unittesting")
def test_is_reference_endpoint_push_scoped_symbol() -> None:
    graph = SyntaxGraph()
    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.SCOPE.value)
    graph.update_node_attribute("2", STACK_GRAPH_IS_EXPORTED, value=True)

    graph.add_node("3")
    graph.update_node_attribute("3", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SCOPED_SYMBOL.value)
    graph.update_node_attribute("3", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("3", STACK_GRAPH_SCOPE, "2")
    graph.update_node_attribute("3", STACK_GRAPH_IS_REFERENCE, value=True)

    view = view_for_graph(graph)
    node_index = view.nid_to_index["3"]

    assert is_reference_endpoint(view, node_index)


@pytest.mark.blends_test_group("stack_unittesting")
def test_is_definition_endpoint_pop_with_flag_true() -> None:
    graph = SyntaxGraph()
    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SYMBOL.value)
    graph.update_node_attribute("2", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("2", STACK_GRAPH_IS_DEFINITION, value=True)

    view = view_for_graph(graph)
    node_index = view.nid_to_index["2"]

    assert is_definition_endpoint(view, node_index)


@pytest.mark.blends_test_group("stack_unittesting")
def test_is_definition_endpoint_pop_with_flag_false() -> None:
    graph = SyntaxGraph()
    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SYMBOL.value)
    graph.update_node_attribute("2", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("2", STACK_GRAPH_IS_DEFINITION, value=False)

    view = view_for_graph(graph)
    node_index = view.nid_to_index["2"]

    assert not is_definition_endpoint(view, node_index)
    assert not is_endpoint(view, node_index)


@pytest.mark.blends_test_group("stack_unittesting")
def test_is_exported_scope_endpoint_exported_scope() -> None:
    graph = SyntaxGraph()
    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.SCOPE.value)
    graph.update_node_attribute("2", STACK_GRAPH_IS_EXPORTED, value=True)

    view = view_for_graph(graph)
    node_index = view.nid_to_index["2"]

    assert is_exported_scope_endpoint(view, node_index)
    assert is_endpoint(view, node_index)


@pytest.mark.blends_test_group("stack_unittesting")
def test_is_exported_scope_endpoint_non_exported_scope() -> None:
    graph = SyntaxGraph()
    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.SCOPE.value)
    graph.update_node_attribute("2", STACK_GRAPH_IS_EXPORTED, value=False)

    view = view_for_graph(graph)
    node_index = view.nid_to_index["2"]

    assert not is_exported_scope_endpoint(view, node_index)
    assert not is_endpoint(view, node_index)


@pytest.mark.blends_test_group("stack_unittesting")
def test_is_root_endpoint() -> None:
    graph = SyntaxGraph()

    view = view_for_graph(graph)
    root_index = 0

    assert is_root_endpoint(view, root_index)
    assert is_endpoint(view, root_index)


@pytest.mark.blends_test_group("stack_unittesting")
def test_is_jump_to_boundary() -> None:
    graph = SyntaxGraph()

    view = view_for_graph(graph)
    jump_to_index = 1

    assert is_jump_to_boundary(view, jump_to_index)
    assert not is_endpoint(view, jump_to_index)


@pytest.mark.blends_test_group("stack_unittesting")
def test_is_complete_definition_path_valid_ref_to_def() -> None:
    graph = SyntaxGraph()
    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SYMBOL.value)
    graph.update_node_attribute("2", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("2", STACK_GRAPH_IS_REFERENCE, value=True)

    graph.add_node("3")
    graph.update_node_attribute("3", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SYMBOL.value)
    graph.update_node_attribute("3", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("3", STACK_GRAPH_IS_DEFINITION, value=True)

    view = view_for_graph(graph)
    start_index = view.nid_to_index["2"]
    end_index = view.nid_to_index["3"]

    start_state = StackState(symbol_stack=None, scope_stack=None)
    result = apply_node(view, start_index, start_state)
    end_state = apply_node(view, end_index, result.state).state

    assert is_complete_definition_path(
        view,
        start_node_index=start_index,
        end_node_index=end_index,
        start_state=start_state,
        end_state=end_state,
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_is_complete_definition_path_false_if_start_not_reference_endpoint() -> None:
    graph = SyntaxGraph()
    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SYMBOL.value)
    graph.update_node_attribute("2", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("2", STACK_GRAPH_IS_REFERENCE, value=False)

    graph.add_node("3")
    graph.update_node_attribute("3", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SYMBOL.value)
    graph.update_node_attribute("3", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("3", STACK_GRAPH_IS_DEFINITION, value=True)

    view = view_for_graph(graph)
    start_index = view.nid_to_index["2"]
    end_index = view.nid_to_index["3"]

    start_state = StackState(symbol_stack=None, scope_stack=None)
    end_state = StackState(symbol_stack=None, scope_stack=None)

    assert not is_complete_definition_path(
        view,
        start_node_index=start_index,
        end_node_index=end_index,
        start_state=start_state,
        end_state=end_state,
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_is_complete_definition_path_false_if_end_not_definition_endpoint() -> None:
    graph = SyntaxGraph()
    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SYMBOL.value)
    graph.update_node_attribute("2", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("2", STACK_GRAPH_IS_REFERENCE, value=True)

    graph.add_node("3")
    graph.update_node_attribute("3", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SYMBOL.value)
    graph.update_node_attribute("3", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("3", STACK_GRAPH_IS_DEFINITION, value=False)

    view = view_for_graph(graph)
    start_index = view.nid_to_index["2"]
    end_index = view.nid_to_index["3"]

    start_state = StackState(symbol_stack=None, scope_stack=None)
    end_state = StackState(symbol_stack=None, scope_stack=None)

    assert not is_complete_definition_path(
        view,
        start_node_index=start_index,
        end_node_index=end_index,
        start_state=start_state,
        end_state=end_state,
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_is_complete_definition_path_false_if_start_symbol_stack_non_empty() -> None:
    graph = SyntaxGraph()
    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SYMBOL.value)
    graph.update_node_attribute("2", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("2", STACK_GRAPH_IS_REFERENCE, value=True)

    graph.add_node("3")
    graph.update_node_attribute("3", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SYMBOL.value)
    graph.update_node_attribute("3", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("3", STACK_GRAPH_IS_DEFINITION, value=True)

    view = view_for_graph(graph)
    start_index = view.nid_to_index["2"]
    end_index = view.nid_to_index["3"]

    sym_id = view.symbols.intern("y")
    start_state = StackState(
        symbol_stack=SymbolStackNode(symbol_id=sym_id, scopes=None, tail=None),
        scope_stack=None,
    )
    end_state = StackState(symbol_stack=None, scope_stack=None)

    assert not is_complete_definition_path(
        view,
        start_node_index=start_index,
        end_node_index=end_index,
        start_state=start_state,
        end_state=end_state,
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_is_complete_definition_path_false_if_start_scope_stack_non_empty() -> None:
    graph = SyntaxGraph()
    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SYMBOL.value)
    graph.update_node_attribute("2", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("2", STACK_GRAPH_IS_REFERENCE, value=True)

    graph.add_node("3")
    graph.update_node_attribute("3", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SYMBOL.value)
    graph.update_node_attribute("3", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("3", STACK_GRAPH_IS_DEFINITION, value=True)

    view = view_for_graph(graph)
    start_index = view.nid_to_index["2"]
    end_index = view.nid_to_index["3"]

    start_state = StackState(
        symbol_stack=None,
        scope_stack=ScopeStackNode(scope_index=10, tail=None),
    )
    end_state = StackState(symbol_stack=None, scope_stack=None)

    assert not is_complete_definition_path(
        view,
        start_node_index=start_index,
        end_node_index=end_index,
        start_state=start_state,
        end_state=end_state,
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_is_complete_definition_path_false_if_end_symbol_stack_non_empty() -> None:
    graph = SyntaxGraph()
    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SYMBOL.value)
    graph.update_node_attribute("2", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("2", STACK_GRAPH_IS_REFERENCE, value=True)

    graph.add_node("3")
    graph.update_node_attribute("3", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SYMBOL.value)
    graph.update_node_attribute("3", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("3", STACK_GRAPH_IS_DEFINITION, value=True)

    view = view_for_graph(graph)
    start_index = view.nid_to_index["2"]
    end_index = view.nid_to_index["3"]

    start_state = StackState(symbol_stack=None, scope_stack=None)
    sym_id = view.symbols.intern("y")
    end_state = StackState(
        symbol_stack=SymbolStackNode(symbol_id=sym_id, scopes=None, tail=None),
        scope_stack=None,
    )

    assert not is_complete_definition_path(
        view,
        start_node_index=start_index,
        end_node_index=end_index,
        start_state=start_state,
        end_state=end_state,
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_is_complete_definition_path_false_if_end_scope_stack_non_empty() -> None:
    graph = SyntaxGraph()
    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SYMBOL.value)
    graph.update_node_attribute("2", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("2", STACK_GRAPH_IS_REFERENCE, value=True)

    graph.add_node("3")
    graph.update_node_attribute("3", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SYMBOL.value)
    graph.update_node_attribute("3", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("3", STACK_GRAPH_IS_DEFINITION, value=True)

    view = view_for_graph(graph)
    start_index = view.nid_to_index["2"]
    end_index = view.nid_to_index["3"]

    start_state = StackState(symbol_stack=None, scope_stack=None)
    result = apply_node(view, start_index, start_state)
    end_result = apply_node(view, end_index, result.state)
    end_state = end_result.state

    end_state_with_scope = StackState(
        symbol_stack=end_state.symbol_stack,
        scope_stack=ScopeStackNode(scope_index=10, tail=None),
    )

    assert not is_complete_definition_path(
        view,
        start_node_index=start_index,
        end_node_index=end_index,
        start_state=start_state,
        end_state=end_state_with_scope,
    )



from blends.symbolic_evaluation.dispatcher import (
    generic,
)
from blends.symbolic_evaluation.models import (
    SymbolicEvalArgs,
    SymbolicEvaluation,
)


def evaluate(args: SymbolicEvalArgs) -> SymbolicEvaluation:
    args.evaluation[args.n_id] = False
    danger_params = False
    if pl_id := args.graph.nodes[args.n_id].get("parameters_id"):
        danger_params = generic(args.fork_n_id(pl_id)).danger
    danger_block = False
    if block_id := args.graph.nodes[args.n_id].get("block_id"):
        danger_block = generic(args.fork_n_id(block_id)).danger

    args.evaluation[args.n_id] = danger_params or danger_block
    if args.method_evaluators and (
        method_evaluator := args.method_evaluators.get("method_declaration")
    ):
        args.evaluation[args.n_id] = method_evaluator(args).danger

    return SymbolicEvaluation(args.evaluation[args.n_id], args.triggers)



from blends.models import (
    Graph,
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_d,
    match_ast_group_d,
    pred,
)
from blends.syntax.builders.method_invocation import (
    DirectChildren,
    build_method_invocation_node,
)
from blends.syntax.builders.symbol_lookup import (
    build_symbol_lookup_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)

SELECTORS_FOR_THREE_PART_INVOCATION = 3
SELECTORS_FOR_TWO_PART_INVOCATION = 2


def _try_build_chained_method_invocation(
    args: SyntaxGraphArgs,
    symbol: str,
    selectors_ids: list[NId],
) -> NId | None:
    graph = args.ast_graph
    child_s1 = None
    child_s2 = None

    if (
        len(selectors_ids) == SELECTORS_FOR_THREE_PART_INVOCATION
        and (child_s0 := match_ast_d(graph, selectors_ids[0], "argument_part"))
        and (args_id := match_ast_d(graph, child_s0, "arguments"))
    ):
        if not match_ast_d(graph, args_id, "(") and not match_ast_d(graph, args_id, ")"):
            return build_symbol_lookup_node(args, symbol)
        child_s1 = match_ast_d(graph, selectors_ids[1], "unconditional_assignable_selector")
        child_s2 = match_ast_d(graph, selectors_ids[2], "argument_part")

    if len(selectors_ids) == SELECTORS_FOR_TWO_PART_INVOCATION:
        child_s1 = match_ast_d(graph, selectors_ids[0], "unconditional_assignable_selector")
        child_s2 = match_ast_d(graph, selectors_ids[1], "argument_part")

    if child_s1 and child_s2 and (expr_id := adj_ast(graph, child_s1)) and len(expr_id) > 1:
        expr = symbol + "." + node_to_str(graph, expr_id[1])
        args_id = match_ast_d(graph, child_s2, "arguments")
        children: DirectChildren = {"expression_id": expr_id[1]}
        if args_id:
            children["arguments_id"] = args_id
        return build_method_invocation_node(args, expr, children)

    return None


def _try_build_single_selector_invocation(
    args: SyntaxGraphArgs,
    graph: Graph,
    symbol: str,
    selectors_ids: list[NId],
) -> NId | None:
    if len(selectors_ids) != 1:
        return None

    child_args = match_ast_d(graph, selectors_ids[0], "argument_part")
    if child_args:
        args_id = match_ast_d(graph, child_args, "arguments")
        children: DirectChildren = {}
        if args_id:
            children["arguments_id"] = args_id
        return build_method_invocation_node(args, symbol, children)

    return None


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    symbol = graph.nodes[args.n_id]["label_text"]
    pred_nid = pred(graph, args.n_id)[0]
    label_type = graph.nodes[pred_nid]["label_type"]

    selectors_ids = match_ast_group_d(graph, pred_nid, "selector")
    if not selectors_ids:
        return build_symbol_lookup_node(args, symbol)

    if label_type in {"initialized_variable_definition", "assignment_expression"} and (
        result := _try_build_chained_method_invocation(args, symbol, selectors_ids)
    ):
        return result

    if label_type in {"static_final_declaration", "initialized_variable_definition"} and (
        result := _try_build_single_selector_invocation(args, graph, symbol, selectors_ids)
    ):
        return result

    return build_symbol_lookup_node(args, symbol)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_d,
    match_ast_group_d,
)
from blends.syntax.builders.literal import (
    build_literal_node,
)
from blends.syntax.builders.object import (
    build_object_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)

GENERATOR_DATA_SOURCES = [
    "aws_secretsmanager_random_password",
    "aws_iam_policy_document",
]


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph

    identifiers_id = match_ast_group_d(graph, args.n_id, "string_lit")
    name = ""
    tf_reference = None
    expected_identifiers_for_tf_reference = 2
    if len(identifiers_id) == expected_identifiers_for_tf_reference:
        name = node_to_str(graph, identifiers_id[0])[1:-1]
        reference = node_to_str(graph, identifiers_id[1])[1:-1]
        tf_reference = f"{name}.{reference}"
    elif identifier := match_ast_d(graph, args.n_id, "identifier"):
        if graph.nodes[identifier].get("label_text") == "dynamic" and len(identifiers_id) >= 1:
            name = node_to_str(graph, identifiers_id[0])[1:-1]
        else:
            name = node_to_str(graph, identifier)

    if (
        (identifier := match_ast_d(graph, args.n_id, "identifier"))
        and graph.nodes[identifier].get("label_text") == "data"
        and name not in GENERATOR_DATA_SOURCES
    ):
        return build_literal_node(args, name, "DataResource")

    if body_id := match_ast_d(graph, args.n_id, "body"):
        c_ids = adj_ast(graph, body_id)
    else:
        c_ids = adj_ast(graph, args.n_id)

    valid_parameters = {
        "attribute",
        "block",
    }

    return build_object_node(
        args,
        c_ids=(_id for _id in c_ids if graph.nodes[_id]["label_type"] in valid_parameters),
        name=name,
        tf_reference=tf_reference,
    )



from blends.models import (
    NId,
)
from blends.query import (
    match_ast_d,
)
from blends.syntax.builders.literal import (
    build_literal_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    value = match_ast_d(graph, args.n_id, "true") or match_ast_d(graph, args.n_id, "false")
    value_text = graph.nodes[value]["label_text"] if value else ""
    return build_literal_node(args, value_text, "bool")



from blends.models import (
    Graph,
    NId,
)
from blends.query import (
    adj_ast,
    get_node_by_path,
    match_ast_d,
    match_ast_group,
)
from blends.syntax.builders.class_decl import (
    build_class_node,
)
from blends.syntax.metadata.java import (
    add_class_to_metadata,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def extract_modifiers(graph: Graph, n_id: NId) -> tuple[NId | None, str | None]:
    modifiers_id = match_ast_d(graph, n_id, "modifiers")
    if not modifiers_id:
        return None, None

    access_modifiers: str | None = None
    annotation_ids_raw = match_ast_group(graph, modifiers_id, "annotation", "marker_annotation")
    annotation_ids = annotation_ids_raw["annotation"] + annotation_ids_raw["marker_annotation"]

    keyword_ids = [c for c in adj_ast(graph, modifiers_id) if c not in annotation_ids]
    if len(annotation_ids) == 0:
        modifiers_id = None

    if keyword_ids:
        access_modifiers = " ".join(graph.nodes[c].get("label_text", "") for c in keyword_ids)
    return modifiers_id, access_modifiers


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    class_node = args.ast_graph.nodes[args.n_id]
    name_id = class_node["label_field_name"]
    block_id = class_node["label_field_body"]
    name = node_to_str(graph, name_id)

    inherited_class: str | None = None
    if extends_id := get_node_by_path(graph, args.n_id, "superclass", "type_identifier"):
        inherited_class = graph.nodes[extends_id].get("label_text", None)

    if (
        implements_id := get_node_by_path(
            graph,
            args.n_id,
            "super_interfaces",
            "type_list",
            "type_identifier",
        )
    ) and (implements_value := graph.nodes[implements_id].get("label_text")):
        if inherited_class is not None:
            inherited_class = f"{inherited_class},{implements_value}"
        else:
            inherited_class = implements_value

    if args.syntax_graph.nodes.get(NId(0)):
        add_class_to_metadata(args, name)

    modifiers_id, access_modifiers = extract_modifiers(graph, args.n_id)

    return build_class_node(
        args,
        name,
        block_id,
        None,
        inherited_class,
        modifiers_id,
        access_modifiers,
    )



from blends.models import (
    NId,
)
from blends.query import (
    match_ast,
)
from blends.syntax.builders.method_invocation import (
    DirectChildren,
    build_method_invocation_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    call_node = args.ast_graph.nodes[args.n_id]
    method_id = call_node["label_field_function"]
    expr = node_to_str(args.ast_graph, method_id)

    args_id = call_node["label_field_arguments"]
    if "__0__" not in match_ast(args.ast_graph, args_id, "(", ")"):
        args_id = None

    children: DirectChildren = {"expression_id": method_id}
    if args_id:
        children["arguments_id"] = args_id

    return build_method_invocation_node(args, expr, children)



from blends.models import (
    NId,
)
from blends.query import (
    match_ast_d,
)
from blends.syntax.builders.variable_declaration import (
    build_variable_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    declarator_id = match_ast_d(graph, args.n_id, "var_spec") or match_ast_d(
        graph,
        args.n_id,
        "const_spec",
    )

    if not declarator_id:
        var_name = node_to_str(graph, args.n_id)
        return build_variable_declaration_node(args, var_name, None, None)

    var_id = graph.nodes[declarator_id]["label_field_name"]
    var_name = node_to_str(graph, var_id)
    type_name = None
    if type_id := graph.nodes[declarator_id].get("label_field_type"):
        type_name = node_to_str(graph, type_id)
    value_id = graph.nodes[declarator_id].get("label_field_value")

    return build_variable_declaration_node(args, var_name, type_name, value_id)



from blends.models import (
    NId,
)
from blends.syntax.builders.comment import (
    build_comment_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    comment = (
        comment_text
        if (comment_text := args.ast_graph.nodes[args.n_id].get("label_text"))
        else node_to_str(args.ast_graph, args.n_id)
    )

    return build_comment_node(args, comment)



from blends.syntax.models import (
    Dispatcher,
)
from blends.syntax.readers.go import (
    argument_list as go_argument_list,
)
from blends.syntax.readers.go import (
    assignment_statement as go_assignment_statement,
)
from blends.syntax.readers.go import (
    binary_expression as go_binary_expression,
)
from blends.syntax.readers.go import (
    block as go_block,
)
from blends.syntax.readers.go import (
    boolean_literal as go_boolean_literal,
)
from blends.syntax.readers.go import (
    break_statement as go_break_statement,
)
from blends.syntax.readers.go import (
    call_expression as go_call_expression,
)
from blends.syntax.readers.go import (
    comment as go_comment,
)
from blends.syntax.readers.go import (
    composite_literal as go_composite_literal,
)
from blends.syntax.readers.go import (
    expression_list as go_expression_list,
)
from blends.syntax.readers.go import (
    expression_statement as go_expression_statement,
)
from blends.syntax.readers.go import (
    for_statement as go_for_statement,
)
from blends.syntax.readers.go import (
    function_declaration as go_function_declaration,
)
from blends.syntax.readers.go import (
    identifier as go_identifier,
)
from blends.syntax.readers.go import (
    if_statement as go_if_statement,
)
from blends.syntax.readers.go import (
    import_declaration as go_import_declaration,
)
from blends.syntax.readers.go import (
    index_expression as go_index_expression,
)
from blends.syntax.readers.go import (
    int_literal as go_int_literal,
)
from blends.syntax.readers.go import (
    literal_value as go_literal_value,
)
from blends.syntax.readers.go import (
    nil as go_nil,
)
from blends.syntax.readers.go import (
    package_clause as go_package_clause,
)
from blends.syntax.readers.go import (
    pair as go_pair,
)
from blends.syntax.readers.go import (
    parameter_declaration as go_parameter_declaration,
)
from blends.syntax.readers.go import (
    parameter_list as go_parameter_list,
)
from blends.syntax.readers.go import (
    qualified_type as go_qualified_type,
)
from blends.syntax.readers.go import (
    reserved_words as go_reserved_words,
)
from blends.syntax.readers.go import (
    return_statement as go_return_statement,
)
from blends.syntax.readers.go import (
    selector_expression as go_selector_expression,
)
from blends.syntax.readers.go import (
    slice_expression as go_slice_expression,
)
from blends.syntax.readers.go import (
    source_file as go_source_file,
)
from blends.syntax.readers.go import (
    spread_element as go_spread_element,
)
from blends.syntax.readers.go import (
    string_literal as go_string_literal,
)
from blends.syntax.readers.go import (
    switch_section as go_switch_section,
)
from blends.syntax.readers.go import (
    switch_statement as go_switch_statement,
)
from blends.syntax.readers.go import (
    type_conversion as go_type_conversion,
)
from blends.syntax.readers.go import (
    type_declaration as go_type_declaration,
)
from blends.syntax.readers.go import (
    unary_expression as go_unary_expression,
)
from blends.syntax.readers.go import (
    var_declaration as go_var_declaration,
)

GO_DISPATCHERS: Dispatcher = {
    "argument_list": go_argument_list.reader,
    "assignment_statement": go_assignment_statement.reader,
    "short_var_declaration": go_assignment_statement.reader,
    "block": go_block.reader,
    "binary_expression": go_binary_expression.reader,
    "break_statement": go_break_statement.reader,
    "false": go_boolean_literal.reader,
    "true": go_boolean_literal.reader,
    "call_expression": go_call_expression.reader,
    "comment": go_comment.reader,
    "composite_literal": go_composite_literal.reader,
    "expression_list": go_expression_list.reader,
    "expression_case": go_switch_section.reader,
    "default_case": go_switch_section.reader,
    "expression_statement": go_expression_statement.reader,
    "for_statement": go_for_statement.reader,
    "func_literal": go_function_declaration.reader,
    "function_declaration": go_function_declaration.reader,
    "method_declaration": go_function_declaration.reader,
    "blank_identifier": go_identifier.reader,
    "identifier": go_identifier.reader,
    "field_identifier": go_identifier.reader,
    "package_identifier": go_identifier.reader,
    "type_identifier": go_identifier.reader,
    "literal_value": go_literal_value.reader,
    "if_statement": go_if_statement.reader,
    "import_spec": go_import_declaration.reader,
    "index_expression": go_index_expression.reader,
    "int_literal": go_int_literal.reader,
    "nil": go_nil.reader,
    "package_clause": go_package_clause.reader,
    "parameter_declaration": go_parameter_declaration.reader,
    "parameter_list": go_parameter_list.reader,
    "qualified_type": go_qualified_type.reader,
    "defer_statement": go_reserved_words.reader,
    "return_statement": go_return_statement.reader,
    "selector_expression": go_selector_expression.reader,
    "slice_expression": go_slice_expression.reader,
    "source_file": go_source_file.reader,
    "interpreted_string_literal": go_string_literal.reader,
    "raw_string_literal": go_string_literal.reader,
    "type_case": go_switch_section.reader,
    "type_declaration": go_type_declaration.reader,
    "expression_switch_statement": go_switch_statement.reader,
    "type_assertion_expression": go_unary_expression.reader,
    "type_switch_statement": go_switch_statement.reader,
    "type_conversion_expression": go_type_conversion.reader,
    "unary_expression": go_unary_expression.reader,
    "const_declaration": go_var_declaration.reader,
    "var_declaration": go_var_declaration.reader,
    "variadic_argument": go_spread_element.reader,
    "keyed_element": go_pair.reader,
}



from blends.models import (
    NId,
)
from blends.query import (
    match_ast_group_d,
)
from blends.syntax.builders.expression_statement import (
    build_expression_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    var_ids = match_ast_group_d(args.ast_graph, args.n_id, "identifier")
    return build_expression_statement_node(args, iter(var_ids))



from blends.models import (
    NId,
)
from blends.query import (
    match_ast_group_d,
)
from blends.syntax.builders.try_statement import (
    build_try_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    block_id = graph.nodes[args.n_id]["label_field_body"]
    catch_blocks = match_ast_group_d(graph, args.n_id, "label_field_handler")
    return build_try_statement_node(args, block_id, catch_blocks, None, None)



from unittest.mock import (
    patch,
)

import pytest

from blends.models import (
    Graph,
    Language,
    NId,
    SyntaxGraph,
)
from blends.syntax.builders.symbol_lookup import (
    build_symbol_lookup_node,
)
from blends.syntax.builders.variable_declaration import (
    build_variable_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
    SyntaxMetadata,
)


@pytest.fixture  # type: ignore[misc,attr-defined]
def mock_args() -> SyntaxGraphArgs:  # type: ignore[misc]
    syntax_graph = SyntaxGraph()
    ast_graph = Graph()
    metadata: SyntaxMetadata = {
        "class_path": [],
        "scope_stack": ["1"],
    }
    syntax_graph.add_node("1", label_type="File")

    def generic(args: SyntaxGraphArgs) -> NId:
        return args.n_id

    return SyntaxGraphArgs(
        generic=generic,
        path="test.py",
        language=Language.PYTHON,
        ast_graph=ast_graph,
        syntax_graph=syntax_graph,
        n_id="2",
        metadata=metadata,
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_symbol_lookup_push_edge_created(
    mock_args: SyntaxGraphArgs,
) -> None:
    with patch("blends.ctx.ctx.has_feature_flag", side_effect=lambda flag: flag == "StackGraph"):
        build_symbol_lookup_node(mock_args, symbol="x")

        assert mock_args.syntax_graph.has_edge("2", "1")
        edge_data = mock_args.syntax_graph["2"]["1"]
        assert edge_data["precedence"] == 0


@pytest.mark.blends_test_group("stack_unittesting")
def test_symbol_lookup_no_edge_when_feature_preview_disabled(
    mock_args: SyntaxGraphArgs,
) -> None:
    with patch("blends.ctx.ctx.has_feature_flag", return_value=False):
        build_symbol_lookup_node(mock_args, symbol="x")

        assert not mock_args.syntax_graph.has_edge("2", "1")


@pytest.mark.blends_test_group("stack_unittesting")
def test_symbol_lookup_no_edge_when_scope_stack_empty(
    mock_args: SyntaxGraphArgs,
) -> None:
    mock_args.metadata["scope_stack"] = []
    with patch("blends.ctx.ctx.has_feature_flag", side_effect=lambda flag: flag == "StackGraph"):
        build_symbol_lookup_node(mock_args, symbol="x")

        assert not mock_args.syntax_graph.has_edge("2", "1")


@pytest.mark.blends_test_group("stack_unittesting")
def test_symbol_lookup_push_edge_in_method_scope(
    mock_args: SyntaxGraphArgs,
) -> None:
    mock_args.metadata["scope_stack"] = ["1", "7"]
    mock_args.syntax_graph.add_node("7", label_type="MethodDeclaration", name="foo")
    with patch("blends.ctx.ctx.has_feature_flag", side_effect=lambda flag: flag == "StackGraph"):
        build_symbol_lookup_node(mock_args, symbol="local_var")

        assert mock_args.syntax_graph.has_edge("2", "7")
        edge_data = mock_args.syntax_graph["2"]["7"]
        assert edge_data["precedence"] == 0


@pytest.mark.blends_test_group("stack_unittesting")
def test_symbol_lookup_to_variable_declaration_complete_path(
    mock_args: SyntaxGraphArgs,
) -> None:
    with patch("blends.ctx.ctx.has_feature_flag", side_effect=lambda flag: flag == "StackGraph"):
        variable_args = SyntaxGraphArgs(
            generic=mock_args.generic,
            path=mock_args.path,
            language=mock_args.language,
            ast_graph=mock_args.ast_graph,
            syntax_graph=mock_args.syntax_graph,
            n_id="2",
            metadata=mock_args.metadata,
        )
        build_variable_declaration_node(
            variable_args,
            variable_name="x",
            variable_type=None,
            value_id=None,
        )

        lookup_args = SyntaxGraphArgs(
            generic=mock_args.generic,
            path=mock_args.path,
            language=mock_args.language,
            ast_graph=mock_args.ast_graph,
            syntax_graph=mock_args.syntax_graph,
            n_id="3",
            metadata=mock_args.metadata,
        )
        build_symbol_lookup_node(lookup_args, symbol="x")

        assert mock_args.syntax_graph.has_edge("3", "1")
        push_edge_data = mock_args.syntax_graph["3"]["1"]
        assert push_edge_data["precedence"] == 0

        assert mock_args.syntax_graph.has_edge("1", "2")
        pop_edge_data = mock_args.syntax_graph["1"]["2"]
        assert pop_edge_data["precedence"] == 1



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.unary_expression import (
    build_unary_expression_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    c_ids = adj_ast(graph, args.n_id)
    val_id = c_ids[0]

    if graph.nodes[val_id]["label_type"] == "assignable_expression":
        val_id = adj_ast(graph, val_id)[0]

    operator = graph.nodes[c_ids[1]].get("label_text", "") if len(c_ids) > 1 else "Unary"

    return build_unary_expression_node(args, operator, val_id)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.member_access import (
    build_member_access_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    c_ids = adj_ast(args.ast_graph, args.n_id)

    obj_id = c_ids[0]
    expression: str = node_to_str(graph, obj_id)

    member_n_id = c_ids[-1]
    member: str = graph.nodes[member_n_id].get("label_text", "")

    return build_member_access_node(args, member, expression, obj_id)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.execution_block import (
    build_execution_block_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    _, *c_ids, _ = adj_ast(args.ast_graph, args.n_id)
    return build_execution_block_node(args, c_ids=(_id for _id in c_ids))



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_group_d,
)
from blends.syntax.builders.string_literal import (
    build_string_literal_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    text = graph.nodes[args.n_id].get("label_text") or node_to_str(graph, args.n_id)
    ints = match_ast_group_d(graph, args.n_id, "template_substitution")
    int_subs = [adj_ast(graph, _id)[1] for _id in ints]
    if len(int_subs) > 0:
        return build_string_literal_node(args, text, iter(int_subs))

    return build_string_literal_node(args, text)



from blends.ctx import ctx
from blends.models import (
    NId,
)
from blends.syntax.builders.assignment import (
    build_assignment_node,
)
from blends.syntax.builders.variable_declaration import (
    build_variable_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.syntax.node_attributes import (
    STACK_GRAPH_IMPORT_MODULE_STEM,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def _find_import_binding(
    args: SyntaxGraphArgs,
    rhs_name: str,
) -> tuple[str, str] | None:
    scope_stack = args.metadata.get("scope_stack", [])
    if not scope_stack:
        return None
    parent_scope = scope_stack[-1]
    for neighbor in args.syntax_graph.successors(parent_scope):
        sg_node = args.syntax_graph.nodes[neighbor]
        if sg_node.get("label_type") != "Import":
            continue
        expression = str(sg_node.get("expression", ""))
        label_alias = sg_node.get("label_alias") or ""
        bound = str(label_alias) if label_alias else expression.split(".")[-1]
        if bound != rhs_name:
            continue
        stem = sg_node.get(STACK_GRAPH_IMPORT_MODULE_STEM)
        if stem:
            return str(stem), expression.split(".")[-1]
    return None


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]
    var_id = n_attrs["label_field_left"]
    val_id = n_attrs.get("label_field_right")
    type_id = n_attrs.get("label_field_type")

    if graph.nodes[var_id]["label_type"] == "identifier" or not val_id:
        var_name = node_to_str(graph, n_attrs["label_field_left"])
        var_type = None
        if type_id:
            var_type = node_to_str(graph, type_id)
        import_module_stem: str | None = None
        import_source_symbol: str | None = None
        if (
            ctx.has_feature_flag("StackGraph")
            and val_id is not None
            and graph.nodes[val_id].get("label_type") == "identifier"
        ):
            rhs_name = node_to_str(graph, val_id)
            result = _find_import_binding(args, rhs_name)
            if result is not None:
                import_module_stem, import_source_symbol = result
        return build_variable_declaration_node(
            args,
            var_name,
            var_type,
            val_id,
            import_module_stem=import_module_stem,
            import_source_symbol=import_source_symbol,
        )

    return build_assignment_node(args, var_id, val_id, type_id)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_d,
    pred_ast,
)
from blends.syntax.builders.parameter import (
    build_parameter_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    val_id = match_ast_d(graph, args.n_id, "identifier")
    var_name = node_to_str(graph, val_id) if val_id else "UnnamedParam"

    var_type_id = match_ast_d(graph, args.n_id, "user_type")
    var_type = node_to_str(graph, var_type_id) if var_type_id else None
    parent_id = pred_ast(graph, args.n_id)[0]

    parent_children = adj_ast(graph, parent_id)
    param_modifier = None

    if (
        args.n_id in parent_children
        and (idx := parent_children.index(args.n_id)) > 0
        and (prev := parent_children[idx - 1])
        and graph.nodes[prev]["label_type"] == "parameter_modifiers"
    ):
        param_modifier = node_to_str(graph, prev)

    return build_parameter_node(
        args=args,
        variable=var_name,
        variable_type=var_type,
        value_id=val_id,
        modifier=param_modifier,
    )



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.selector import (
    build_selector_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph

    c_ids = [
        child
        for child in adj_ast(graph, args.n_id)
        if args.ast_graph.nodes[child]["label_type"] in {"identifier", "index_selector"}
    ]

    return build_selector_node(args, None, iter(c_ids))



from blends.ctx import ctx
from blends.models import (
    NId,
)
from blends.stack.edges import (
    Edge,
    add_edge,
)
from blends.stack.node_helpers import (
    import_binding_node_attributes,
)
from blends.syntax.builders.import_reference_chain import (
    build_import_reference_chain_to_root,
    extract_module_stem,
)
from blends.syntax.builders.utils import (
    get_next_binding_precedence,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def _bound_import_symbol(*, expression: str, alias: str | None) -> str:
    if isinstance(alias, str) and alias:
        return alias
    parts = expression.split(".")
    return parts[-1] if parts else expression


def _mark_import_as_definition(
    args: SyntaxGraphArgs, *, node_id: NId, expression: str, alias: str | None
) -> None:
    if not ctx.has_feature_flag("StackGraph"):
        return
    bound = _bound_import_symbol(expression=expression, alias=alias)
    if not bound:
        return
    precedence = get_next_binding_precedence(args)
    module_stem = extract_module_stem(expression)
    source_symbol = expression.split(".")[-1] if alias else None
    args.syntax_graph.update_node(
        node_id,
        import_binding_node_attributes(
            symbol=bound,
            precedence=precedence,
            import_module_stem=module_stem,
            import_source_symbol=source_symbol,
        ),
    )
    scope_stack = args.metadata.setdefault("scope_stack", [])
    if scope_stack and (parent_scope := scope_stack[-1]):
        add_edge(
            args.syntax_graph,
            Edge(source=parent_scope, sink=node_id, precedence=precedence),
        )
    build_import_reference_chain_to_root(args, module_stem, node_id)


def build_import_statement_node(args: SyntaxGraphArgs, *imported_elements: dict[str, str]) -> NId:
    if len(imported_elements) == 1:
        imported_elements[0].pop("corrected_n_id", None)
        args.syntax_graph.add_node(
            args.n_id,
            **imported_elements[0],
            label_type="Import",
        )
        expression = imported_elements[0].get("expression", "")
        alias = imported_elements[0].get("label_alias")
        _mark_import_as_definition(args, node_id=args.n_id, expression=expression, alias=alias)
    else:
        args.syntax_graph.add_node(
            args.n_id,
            import_type="multiple_import",
            label_type="Import",
        )
        for imported_element in imported_elements:
            corrected_n_id = imported_element.pop("corrected_n_id", None)
            imported_element.update({"label_type": "Import"})
            if corrected_n_id:
                args.syntax_graph.add_node(NId(corrected_n_id), **imported_element)
                args.syntax_graph.add_edge(
                    args.n_id,
                    NId(corrected_n_id),
                    label_ast="AST",
                )
                expression = imported_element.get("expression", "")
                alias = imported_element.get("label_alias")
                _mark_import_as_definition(
                    args,
                    node_id=NId(corrected_n_id),
                    expression=expression,
                    alias=alias,
                )

    return args.n_id



from typing import TypedDict, cast

from blends.models import (
    NId,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


class DirectChildren(TypedDict, total=False):
    expression_id: NId
    arguments_id: NId
    object_id: NId
    block_id: NId


def _process_invocation_children(
    args: SyntaxGraphArgs,
    children: DirectChildren,
) -> None:
    for name_node, child_id in cast("dict[str, NId]", children).items():
        args.syntax_graph.update_node_attribute(args.n_id, name_node, child_id)
        args.syntax_graph.add_edge(
            args.n_id,
            args.generic(args.fork_n_id(child_id)),
            label_ast="AST",
        )


def build_method_invocation_node(
    args: SyntaxGraphArgs,
    expr: str,
    direct_children: DirectChildren,
    object_name: str | None = None,
) -> NId:
    args.syntax_graph.add_node(
        args.n_id,
        expression=expr,
        label_type="MethodInvocation",
    )
    if object_name:
        args.syntax_graph.update_node_attribute(args.n_id, "object", object_name)
    _process_invocation_children(args, direct_children)
    return args.n_id



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast,
    match_ast_d,
)
from blends.syntax.builders.string_literal import (
    build_string_literal_node,
)
from blends.syntax.builders.variable_declaration import (
    build_variable_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    definition_id = match_ast_d(graph, args.n_id, "initialized_variable_definition")
    if not definition_id:
        return build_string_literal_node(args, args.n_id)

    var_type = (
        fr_child
        if (fr_child := match_ast(graph, args.n_id).get("__0__"))
        and graph.nodes[fr_child]["label_type"] == "type_identifier"
        else None
    )

    var_id = graph.nodes[definition_id]["label_field_name"]
    var_name = node_to_str(graph, var_id)

    value_id = graph.nodes[definition_id].get("label_field_value")
    if value_id and graph.nodes[value_id]["label_type"] == "identifier":
        c_ids = [
            child
            for child in adj_ast(graph, definition_id)
            if graph.nodes[child]["label_type"] == "selector"
        ]
        for c_id in c_ids:
            args.syntax_graph.add_edge(
                value_id,
                args.generic(args.fork_n_id(c_id)),
                label_ast="AST",
            )

    return build_variable_declaration_node(args, var_name, var_type, value_id)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from blends.stack.criteria import (
    is_complete_definition_path,
    is_reference_endpoint,
)
from blends.stack.partial_path.errors import PartialPathResolutionError
from blends.stack.partial_path.stack_conditions import create_seed_path
from blends.stack.stitcher_utils import (
    _path_priority,
    _should_cancel,
    _stack_state_from_partial,
)

if TYPE_CHECKING:
    from collections.abc import Callable

    from blends.stack.partial_path import (
        FileHandle,
        PartialPath,
        PartialScopeStack,
        PartialSymbolStack,
    )
    from blends.stack.partial_path_db import PartialPathDb
    from blends.stack.view import StackGraphView

from blends.stack.stacks import StackState

StateKey = tuple[int, int, "PartialSymbolStack", "PartialScopeStack"]


@dataclass(frozen=True, slots=True)
class ForwardStitcherConfig:
    max_work_per_phase: int | None = None
    max_phases: int | None = None
    max_total_work: int | None = None
    max_queue_size: int | None = None
    enable_dedupe: bool = True
    enable_cycle_guard: bool = True
    is_cancelled: Callable[[], bool] | None = None
    deadline: float | None = None


@dataclass(slots=True)
class ForwardStitcherStats:
    phases_executed: int = 0
    candidates_considered: int = 0
    concatenations_succeeded: int = 0
    complete_paths_found: int = 0
    cancelled: bool = False
    cycle_guard_hits: int = 0
    queue_pruned_count: int = 0


@dataclass(frozen=True, slots=True)
class ForwardStitcherStatsSnapshot:
    phases_executed: int
    candidates_considered: int
    concatenations_succeeded: int
    complete_paths_found: int
    cancelled: bool
    cycle_guard_hits: int
    queue_pruned_count: int


@dataclass(frozen=True, slots=True)
class ForwardStitcherLoadRequests:
    node_requests: tuple[tuple[FileHandle, int], ...] = ()
    root_requests: tuple[tuple[FileHandle, PartialSymbolStack], ...] = ()


@dataclass(frozen=True, slots=True)
class ForwardStitcherPhaseResult:
    complete_paths: tuple[PartialPath, ...]
    stats: ForwardStitcherStatsSnapshot


@dataclass(slots=True)
class ForwardStitcher:
    view: StackGraphView
    file_handle: FileHandle
    db: PartialPathDb
    config: ForwardStitcherConfig = field(default_factory=ForwardStitcherConfig)

    _current_queue: list[PartialPath] = field(default_factory=list)
    _previous_phase_paths: tuple[PartialPath, ...] = ()
    _stats: ForwardStitcherStats = field(default_factory=ForwardStitcherStats)
    _is_complete: bool = False
    _seen_states: set[StateKey] = field(default_factory=set)

    @classmethod
    def from_references(
        cls,
        *,
        view: StackGraphView,
        file_handle: FileHandle,
        db: PartialPathDb,
        reference_nodes: tuple[int, ...],
        config: ForwardStitcherConfig | None = None,
    ) -> ForwardStitcher:
        seed_paths: list[PartialPath] = []
        for node_index in sorted(reference_nodes):
            if not is_reference_endpoint(view, node_index):
                continue
            seed_paths.append(create_seed_path(view, node_index))

        stitcher = cls(
            view=view,
            file_handle=file_handle,
            db=db,
            config=config or ForwardStitcherConfig(),
        )
        stitcher._current_queue = seed_paths
        stitcher._is_complete = len(seed_paths) == 0

        if stitcher.config.enable_cycle_guard:
            for seed_path in seed_paths:
                state_key: StateKey = (
                    seed_path.start_node_index,
                    seed_path.end_node_index,
                    seed_path.symbol_stack_postcondition,
                    seed_path.scope_stack_postcondition,
                )
                stitcher._seen_states.add(state_key)

        return stitcher

    def process_next_phase(self) -> ForwardStitcherPhaseResult:
        if self._is_complete or self._phase_limit_reached():
            self._is_complete = True
            return self._build_phase_result(())

        self._stats.phases_executed += 1
        self._previous_phase_paths = tuple(self._current_queue)

        complete_paths: list[PartialPath] = []
        self._current_queue = self._expand_phase(self._current_queue, complete_paths)
        if len(self._current_queue) == 0:
            self._is_complete = True

        if self._phase_limit_reached():
            self._is_complete = True

        return self._build_phase_result(tuple(complete_paths))

    def previous_phase_paths(self) -> tuple[PartialPath, ...]:
        return self._previous_phase_paths

    def is_complete(self) -> bool:
        return self._is_complete

    def _phase_limit_reached(self) -> bool:
        return (
            self.config.max_phases is not None
            and self._stats.phases_executed >= self.config.max_phases
        )

    def _build_phase_result(
        self,
        complete_paths: tuple[PartialPath, ...],
    ) -> ForwardStitcherPhaseResult:
        return ForwardStitcherPhaseResult(
            complete_paths=complete_paths,
            stats=_snapshot_stats(self._stats),
        )

    def _expand_phase(
        self,
        current_queue: list[PartialPath],
        complete_paths: list[PartialPath],
    ) -> list[PartialPath]:
        next_queue: list[PartialPath] = []
        work_performed_phase = 0

        for index, current_path in enumerate(current_queue):
            if _should_cancel(self.config):
                self._stats.cancelled = True
                self._is_complete = True
                return []
            if (
                self.config.max_work_per_phase is not None
                and work_performed_phase >= self.config.max_work_per_phase
            ):
                next_queue.extend(current_queue[index:])
                break

            if (
                self.config.max_total_work is not None
                and self._stats.candidates_considered >= self.config.max_total_work
            ):
                self._is_complete = True
                break

            candidates_processed = self._extend_with_candidates(
                current_path, next_queue, complete_paths
            )
            work_performed_phase += candidates_processed

        if self.config.max_queue_size is not None and len(next_queue) > self.config.max_queue_size:
            pruned_count = len(next_queue) - self.config.max_queue_size
            self._stats.queue_pruned_count += pruned_count
            next_queue.sort(key=_path_priority)
            kept = next_queue[: self.config.max_queue_size]
            pruned = next_queue[self.config.max_queue_size :]
            if self.config.enable_cycle_guard:
                for pruned_path in pruned:
                    state_key: StateKey = (
                        pruned_path.start_node_index,
                        pruned_path.end_node_index,
                        pruned_path.symbol_stack_postcondition,
                        pruned_path.scope_stack_postcondition,
                    )
                    self._seen_states.discard(state_key)
            next_queue = kept

        return next_queue

    def _extend_with_candidates(
        self,
        current_path: PartialPath,
        next_queue: list[PartialPath],
        complete_paths: list[PartialPath],
    ) -> int:
        candidates = self.db.get_candidates(
            file_handle=self.file_handle,
            end_node_index=current_path.end_node_index,
            symbol_stack=current_path.symbol_stack_postcondition,
            scope_stack=current_path.scope_stack_postcondition,
        )
        self._stats.candidates_considered += len(candidates)

        for candidate_id in candidates:
            candidate_path = self.db.get_path(candidate_id)
            if candidate_path is None:
                continue

            try:
                extended = current_path.concatenate(self.view, candidate_path)
            except PartialPathResolutionError:
                continue

            if self.config.enable_cycle_guard:
                state_key: StateKey = (
                    extended.start_node_index,
                    extended.end_node_index,
                    extended.symbol_stack_postcondition,
                    extended.scope_stack_postcondition,
                )
                if state_key in self._seen_states:
                    self._stats.cycle_guard_hits += 1
                    continue
                self._seen_states.add(state_key)

            self._stats.concatenations_succeeded += 1

            if _is_complete_path(self.view, extended):
                complete_paths.append(extended)
                self._stats.complete_paths_found += 1
            else:
                next_queue.append(extended)

        return len(candidates)


def _snapshot_stats(stats: ForwardStitcherStats) -> ForwardStitcherStatsSnapshot:
    return ForwardStitcherStatsSnapshot(
        phases_executed=stats.phases_executed,
        candidates_considered=stats.candidates_considered,
        concatenations_succeeded=stats.concatenations_succeeded,
        complete_paths_found=stats.complete_paths_found,
        cancelled=stats.cancelled,
        cycle_guard_hits=stats.cycle_guard_hits,
        queue_pruned_count=stats.queue_pruned_count,
    )


def _is_complete_path(
    view: StackGraphView,
    path: PartialPath,
) -> bool:
    start_state = _stack_state_from_partial(
        path.symbol_stack_precondition, path.scope_stack_precondition
    )
    end_state = _stack_state_from_partial(
        path.symbol_stack_postcondition, path.scope_stack_postcondition
    )

    if end_state.scope_stack is not None and path.scope_stack_postcondition.has_variable():
        end_state = StackState(
            symbol_stack=end_state.symbol_stack,
            scope_stack=None,
        )
    return is_complete_definition_path(
        view,
        start_node_index=path.start_node_index,
        end_node_index=path.end_node_index,
        start_state=start_state,
        end_state=end_state,
    )



from blends.models import (
    NId,
)
from blends.query import (
    pred_ast,
)
from blends.syntax.builders.named_argument import (
    build_named_argument_node,
)
from blends.syntax.builders.pair import (
    build_pair_node,
)
from blends.syntax.builders.symbol_lookup import (
    build_symbol_lookup_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    key_id = args.ast_graph.nodes[args.n_id]["label_field_key"]
    if not (value_id := args.ast_graph.nodes[args.n_id].get("label_field_value")):
        symbol = node_to_str(args.ast_graph, key_id)
        return build_symbol_lookup_node(args, symbol)

    parent = pred_ast(args.ast_graph, args.n_id)
    if args.ast_graph.nodes[parent[0]]["label_type"] == "argument_list":
        arg_name = node_to_str(args.ast_graph, key_id)
        return build_named_argument_node(args, arg_name, value_id)

    return build_pair_node(args, key_id, value_id)



from blends.models import (
    NId,
)
from blends.syntax.builders.method_invocation import (
    DirectChildren,
    build_method_invocation_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    expr_id = graph.nodes[args.n_id]["label_field_function"]
    expr_id_child = None
    children: DirectChildren = {}

    if graph.nodes[expr_id]["label_type"] == "field_expression":
        if object_id := graph.nodes[expr_id].get("label_field_value"):
            children["object_id"] = object_id
        expr = node_to_str(graph, graph.nodes[expr_id]["label_field_field"])
        expr_id_child = graph.nodes[expr_id]["label_field_field"]
    elif graph.nodes[expr_id]["label_type"] == "call_expression":
        children["object_id"] = expr_id
        expr = ""
    else:
        expr = node_to_str(graph, expr_id)

    if expr_id_child:
        children["expression_id"] = expr_id_child

    arguments_id = graph.nodes[args.n_id].get("label_field_arguments")
    if arguments_id and graph.nodes[arguments_id]["label_type"] == "block":
        children["block_id"] = arguments_id
        return build_method_invocation_node(args, expr, children)

    if arguments_id:
        children["arguments_id"] = arguments_id

    return build_method_invocation_node(args, expr, children)



from blends.models import (
    NId,
)
from blends.query import (
    match_ast,
)
from blends.syntax.builders.if_statement import (
    build_if_node,
)
from blends.syntax.builders.string_literal import (
    build_string_literal_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    conditional_node = None
    invalid_types = {
        "?",
        ":",
    }

    true_block = args.ast_graph.nodes[args.n_id]["label_field_consequence"]
    false_block = args.ast_graph.nodes[args.n_id]["label_field_alternative"]

    reserved_ids = {
        true_block,
        false_block,
    }

    first_child = match_ast(args.ast_graph, args.n_id).get("__0__")
    if (
        first_child
        and args.ast_graph.nodes[first_child]["label_type"] not in invalid_types
        and first_child not in reserved_ids
    ):
        conditional_node = first_child

    if not conditional_node:
        return build_string_literal_node(args, node_to_str(args.ast_graph, args.n_id))

    return build_if_node(args, conditional_node, true_block, false_block)



from blends.models import (
    NId,
)
from blends.syntax.builders.member_access import (
    build_member_access_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    expression_id = graph.nodes[args.n_id]["label_field_object"]
    expression = node_to_str(graph, expression_id) if expression_id else "DefaultObject"

    member_id = graph.nodes[args.n_id].get("label_field_property", None)
    member = node_to_str(graph, member_id) if member_id else "DefaultMember"
    return build_member_access_node(args, member, expression, expression_id)



from blends.models import (
    NId,
)
from blends.query import (
    get_node_by_path,
    match_ast,
)
from blends.syntax.builders.variable_declaration import (
    build_variable_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    nodes = graph.nodes

    if name_n_id := get_node_by_path(
        graph,
        args.n_id,
        "property_element",
        "variable_name",
        "name",
    ):
        var_name = nodes[name_n_id].get("label_text", "")
    else:
        var_name = node_to_str(graph, args.n_id)

    val_id: NId | None = None
    if (
        val_p_id := get_node_by_path(
            graph,
            args.n_id,
            "property_element",
            "property_initializer",
        )
    ) and (c_ids := match_ast(graph, val_p_id)):
        val_id = c_ids.get("__1__")

    return build_variable_declaration_node(args, var_name, None, val_id)



from blends.models import (
    NId,
)
from blends.syntax.builders.binary_operation import (
    build_binary_operation_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    left_id = args.ast_graph.nodes[args.n_id]["label_field_left"]
    right_id = args.ast_graph.nodes[args.n_id]["label_field_right"]
    operator_id = args.ast_graph.nodes[args.n_id]["label_field_operator"]
    operator = args.ast_graph.nodes[operator_id]["label_text"]
    return build_binary_operation_node(args, operator, left_id, right_id)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.switch_section import (
    build_switch_section_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    child_ids = adj_ast(graph, args.n_id)
    value_id = child_ids[0]
    case_value = node_to_str(graph, value_id) if value_id else "Default"
    body_id = child_ids[2]
    return build_switch_section_node(args, case_value, [body_id])



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.declaration_block import (
    build_declaration_block_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    c_ids = [
        n_id
        for n_id in adj_ast(args.ast_graph, args.n_id)
        if graph.nodes[n_id]["label_type"] not in {"?", ";", "nullable_type"}
    ]
    if graph.nodes[c_ids[-1]]["label_type"] == "static_final_declaration_list":
        return args.generic(args.fork_n_id(c_ids[-1]))

    return build_declaration_block_node(
        args,
        c_ids=(_id for _id in c_ids if graph.nodes[_id]["label_type"]),
    )



from blends.models import (
    NId,
)
from blends.syntax.builders.parameter import (
    build_parameter_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    var_name = None
    if var_id := graph.nodes[args.n_id].get("label_field_name"):
        var_name = node_to_str(graph, var_id)
    type_id = graph.nodes[args.n_id]["label_field_type"]
    var_type = node_to_str(graph, type_id)

    return build_parameter_node(args=args, variable=var_name, variable_type=var_type, value_id=None)



from collections.abc import (
    Iterator,
)

from blends.query import adj_ast
from blends.symbolic_evaluation.context.search.model import (
    SearchArgs,
    SearchResult,
)


def search(args: SearchArgs) -> Iterator[SearchResult]:
    assign_id = args.graph.nodes[args.n_id]["variable_id"]

    if args.symbol == args.graph.nodes[assign_id].get("symbol"):
        if args.graph.nodes[args.n_id].get("operator") in {"+=", "*=", "/="}:
            if not args.def_only:
                yield False, args.n_id
        else:
            yield True, args.n_id
    elif args.graph.nodes[assign_id].get("label_type") == "ArgumentList":
        graph = args.graph
        for n_id in adj_ast(graph, assign_id, 1, label_type="SymbolLookup"):
            if args.symbol == graph.nodes[n_id].get("symbol"):
                yield True, args.n_id
    elif (
        not args.def_only
        and (expr := args.graph.nodes[assign_id].get("expression"))
        and args.symbol in expr
    ):
        yield False, args.n_id



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.if_statement import (
    build_if_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    condition_id = adj_ast(args.ast_graph, args.n_id)[-1]
    return build_if_node(args, condition_id)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.pair import (
    build_pair_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    key_id = args.ast_graph.nodes[args.n_id]["label_field_key"]
    value_id = args.ast_graph.nodes[args.n_id]["label_field_value"]
    key_child = adj_ast(args.ast_graph, key_id)
    value_child = adj_ast(args.ast_graph, value_id)

    return build_pair_node(args, key_child[0], value_child[0])



from blends.models import (
    NId,
)
from blends.syntax.builders.literal import (
    build_literal_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    n_attrs = args.ast_graph.nodes[args.n_id]
    value = n_attrs.get("label_text") or "boolean"
    return build_literal_node(args, value, "bool")



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.array_node import (
    build_array_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    invalid_types = {"[", "]", "{", "}", ",", ";"}

    valid_seqs = [
        child
        for child in adj_ast(graph, args.n_id)
        if graph.nodes[child]["label_type"] not in invalid_types
    ]

    usable_ids = []
    min_children_for_valid_sequence = 2
    for _id in valid_seqs:
        childs = adj_ast(graph, _id)
        if len(childs) < min_children_for_valid_sequence:
            continue
        current_id = childs[1]

        if graph.nodes[current_id]["label_type"] != "block_node":
            usable_ids.append(current_id)
        elif (current_id_childs := adj_ast(graph, current_id)) and len(current_id_childs) > 0:
            if graph.nodes[current_id_childs[0]]["label_type"] == "tag":
                usable_ids.append(current_id_childs[-1])
            else:
                usable_ids.append(current_id_childs[0])

    return build_array_node(args, usable_ids)



from blends.models import (
    NId,
)
from blends.query import (
    get_nodes_by_path,
    match_ast_group_d,
)
from blends.syntax.builders.class_decl import (
    build_class_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    class_node = args.ast_graph.nodes[args.n_id]
    name_id = class_node["label_field_name"]
    name = node_to_str(args.ast_graph, name_id)
    block_id = class_node.get("label_field_body")
    parameters: list[NId] = []
    annotations = match_ast_group_d(args.ast_graph, args.n_id, "annotation")
    annotation_parameters = get_nodes_by_path(args.ast_graph, args.n_id, "annotation", "arguments")
    class_parameters = match_ast_group_d(args.ast_graph, args.n_id, "class_parameters")
    parameters.extend(annotations)
    parameters.extend(annotation_parameters)
    parameters.extend(class_parameters)

    return build_class_node(args, name, block_id, parameters)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.execution_block import (
    build_execution_block_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    labels_to_ignore = {"else", "do", "end", ";", "{", "}"}
    c_ids = (
        node_id
        for node_id in adj_ast(graph, args.n_id)
        if graph.nodes[node_id]["label_type"] not in labels_to_ignore
    )

    return build_execution_block_node(args, c_ids)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast,
)
from blends.syntax.builders.for_statement import (
    build_for_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph

    c_ids = match_ast(
        graph,
        args.n_id,
        "assignment_expression",
        "binary_expression",
        "update_expression",
    )

    initializer_id = c_ids.get("assignment_expression")
    condition_id = c_ids.get("binary_expression")
    update_id = c_ids.get("update_expression")
    body_id = adj_ast(graph, args.n_id)[-1]

    if graph.nodes[body_id]["label_type"] == "expression_statement":
        body_id = adj_ast(graph, body_id)[0]

    return build_for_statement_node(args, initializer_id, condition_id, update_id, body_id)



from blends.models import (
    NId,
)
from blends.syntax.builders.unary_expression import (
    build_unary_expression_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]
    value_id = n_attrs["label_field_operand"]
    operand = node_to_str(graph, n_attrs["label_field_type"])

    return build_unary_expression_node(args, operand, value_id)



from blends.models import (
    NId,
)
from blends.syntax.builders.import_statement import (
    build_import_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    node_attrs: dict[str, str] = {
        "expression": node_to_str(args.ast_graph, args.n_id),
    }
    return build_import_statement_node(args, node_attrs)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast,
    search_pred_until_type,
)
from blends.syntax.builders.method_declaration import (
    DirectChildren,
    build_method_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    childs = adj_ast(graph, args.n_id)
    if len(childs) == 1 and graph.nodes[childs[0]]["label_type"] == "function_signature":
        return args.generic(args.fork_n_id(childs[0]))

    body_parents = {
        "class_body",
        "extension_body",
    }
    m_name = None
    body_id = None

    parents = search_pred_until_type(graph, args.n_id, body_parents)
    if parents and (class_childs := list(adj_ast(graph, parents[0]))):
        al_list = match_ast(graph, args.n_id, "formal_parameter_list", "static")

        parameters_id = al_list.get("formal_parameter_list")
        body_id = class_childs[class_childs.index(parents[1]) + 1]

        direct: DirectChildren = {}
        if body_id:
            direct["block_id"] = body_id
        if parameters_id:
            direct["parameters_id"] = parameters_id

        return build_method_declaration_node(args, m_name, direct, {})

    return build_method_declaration_node(args, m_name, {}, {})



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from blends.stack.criteria import (
    is_definition_endpoint,
    is_import_binding_endpoint,
    is_reference_endpoint,
)
from blends.stack.forward_stitcher import (
    ForwardStitcherConfig,
    ForwardStitcherLoadRequests,
    ForwardStitcherStats,
    ForwardStitcherStatsSnapshot,
    _snapshot_stats,
)
from blends.stack.partial_path.errors import PartialPathResolutionError
from blends.stack.partial_path.stack_conditions import create_seed_path
from blends.stack.stacks import StackState
from blends.stack.stitcher_utils import (
    _path_priority,
    _should_cancel,
    _stack_state_from_partial,
)

if TYPE_CHECKING:
    from blends.stack.partial_path import (
        FileHandle,
        PartialPath,
        PartialScopeStack,
        PartialSymbolStack,
    )
    from blends.stack.partial_path_db import PartialPathDb
    from blends.stack.view import StackGraphView


MultiFileStateKey = tuple["FileHandle", int, int, "PartialSymbolStack", "PartialScopeStack"]


@dataclass(frozen=True, slots=True)
class MultiFileForwardStitcherPhaseResult:
    complete_paths: tuple[tuple[FileHandle, PartialPath], ...]
    load_requests: ForwardStitcherLoadRequests
    stats: ForwardStitcherStatsSnapshot


@dataclass(slots=True)
class MultiFileForwardStitcher:
    view: StackGraphView
    file_handle: FileHandle
    db: PartialPathDb
    views: dict[FileHandle, StackGraphView] = field(default_factory=dict)
    config: ForwardStitcherConfig = field(default_factory=ForwardStitcherConfig)

    _current_queue: list[tuple[FileHandle, PartialPath]] = field(default_factory=list)
    _previous_phase_paths: tuple[tuple[FileHandle, PartialPath], ...] = ()
    _stats: ForwardStitcherStats = field(default_factory=ForwardStitcherStats)
    _is_complete: bool = False
    _seen_states: set[MultiFileStateKey] = field(default_factory=set)

    @classmethod
    def from_references(  # noqa: PLR0913
        cls,
        *,
        view: StackGraphView,
        file_handle: FileHandle,
        db: PartialPathDb,
        reference_nodes: tuple[int, ...],
        config: ForwardStitcherConfig | None = None,
        views: dict[FileHandle, StackGraphView] | None = None,
    ) -> MultiFileForwardStitcher:
        seed_paths: list[tuple[FileHandle, PartialPath]] = []
        for node_index in sorted(reference_nodes):
            if not is_reference_endpoint(view, node_index):
                continue
            seed_paths.append((file_handle, create_seed_path(view, node_index)))

        stitcher = cls(
            view=view,
            file_handle=file_handle,
            db=db,
            views=views if views is not None else {file_handle: view},
            config=config or ForwardStitcherConfig(),
        )
        stitcher._current_queue = seed_paths
        stitcher._is_complete = len(seed_paths) == 0

        if stitcher.config.enable_cycle_guard:
            for current_fh, seed_path in seed_paths:
                state_key: MultiFileStateKey = (
                    current_fh,
                    seed_path.start_node_index,
                    seed_path.end_node_index,
                    seed_path.symbol_stack_postcondition,
                    seed_path.scope_stack_postcondition,
                )
                stitcher._seen_states.add(state_key)

        return stitcher

    def process_next_phase(self) -> MultiFileForwardStitcherPhaseResult:
        if self._is_complete or self._phase_limit_reached():
            self._is_complete = True
            return self._build_phase_result(ForwardStitcherLoadRequests(), ())

        self._stats.phases_executed += 1
        self._previous_phase_paths = tuple(self._current_queue)

        complete_paths: list[tuple[FileHandle, PartialPath]] = []
        self._current_queue = self._expand_phase(self._current_queue, complete_paths)
        if len(self._current_queue) == 0:
            self._is_complete = True

        load_requests = _collect_load_requests(self._current_queue)
        if self._phase_limit_reached():
            self._is_complete = True

        return self._build_phase_result(load_requests, tuple(complete_paths))

    def previous_phase_paths(self) -> tuple[tuple[FileHandle, PartialPath], ...]:
        return self._previous_phase_paths

    def is_complete(self) -> bool:
        return self._is_complete

    def current_load_requests(self) -> ForwardStitcherLoadRequests:
        return _collect_load_requests(self._current_queue)

    def _phase_limit_reached(self) -> bool:
        return (
            self.config.max_phases is not None
            and self._stats.phases_executed >= self.config.max_phases
        )

    def _build_phase_result(
        self,
        load_requests: ForwardStitcherLoadRequests,
        complete_paths: tuple[tuple[FileHandle, PartialPath], ...],
    ) -> MultiFileForwardStitcherPhaseResult:
        return MultiFileForwardStitcherPhaseResult(
            complete_paths=complete_paths,
            load_requests=load_requests,
            stats=_snapshot_stats(self._stats),
        )

    def _expand_phase(
        self,
        current_queue: list[tuple[FileHandle, PartialPath]],
        complete_paths: list[tuple[FileHandle, PartialPath]],
    ) -> list[tuple[FileHandle, PartialPath]]:
        next_queue: list[tuple[FileHandle, PartialPath]] = []
        work_performed_phase = 0

        for index, (current_fh, current_path) in enumerate(current_queue):
            if _should_cancel(self.config):
                self._stats.cancelled = True
                self._is_complete = True
                return []
            if (
                self.config.max_work_per_phase is not None
                and work_performed_phase >= self.config.max_work_per_phase
            ):
                next_queue.extend(current_queue[index:])
                break
            if (
                self.config.max_total_work is not None
                and self._stats.candidates_considered >= self.config.max_total_work
            ):
                self._is_complete = True
                break
            candidates_processed = self._extend_with_candidates(
                current_fh, current_path, next_queue, complete_paths
            )
            work_performed_phase += candidates_processed

        if self.config.max_queue_size is not None and len(next_queue) > self.config.max_queue_size:
            pruned_count = len(next_queue) - self.config.max_queue_size
            self._stats.queue_pruned_count += pruned_count
            next_queue.sort(key=lambda item: _path_priority(item[1]))
            kept_pairs = next_queue[: self.config.max_queue_size]
            pruned_pairs = next_queue[self.config.max_queue_size :]
            if self.config.enable_cycle_guard:
                for pruned_fh, pruned_path in pruned_pairs:
                    state_key: MultiFileStateKey = (
                        pruned_fh,
                        pruned_path.start_node_index,
                        pruned_path.end_node_index,
                        pruned_path.symbol_stack_postcondition,
                        pruned_path.scope_stack_postcondition,
                    )
                    self._seen_states.discard(state_key)
            next_queue = kept_pairs

        return next_queue

    def _extend_with_candidates(
        self,
        current_fh: FileHandle,
        current_path: PartialPath,
        next_queue: list[tuple[FileHandle, PartialPath]],
        complete_paths: list[tuple[FileHandle, PartialPath]],
    ) -> int:
        candidates = self.db.get_candidates(
            file_handle=current_fh,
            end_node_index=current_path.end_node_index,
            symbol_stack=current_path.symbol_stack_postcondition,
            scope_stack=current_path.scope_stack_postcondition,
        )
        self._stats.candidates_considered += len(candidates)
        current_view = self.views.get(current_fh, self.view)
        for candidate_id in candidates:
            candidate_path = self.db.get_path(candidate_id)
            if candidate_path is None:
                continue
            candidate_fh: FileHandle = candidate_id[0]
            try:
                extended = current_path.concatenate(current_view, candidate_path)
            except PartialPathResolutionError:
                continue
            if self.config.enable_cycle_guard:
                state_key: MultiFileStateKey = (
                    candidate_fh,
                    extended.start_node_index,
                    extended.end_node_index,
                    extended.symbol_stack_postcondition,
                    extended.scope_stack_postcondition,
                )
                if state_key in self._seen_states:
                    self._stats.cycle_guard_hits += 1
                    continue
                self._seen_states.add(state_key)
            self._stats.concatenations_succeeded += 1
            candidate_view = self.views.get(candidate_fh, self.view)
            self._route_extended(candidate_fh, extended, candidate_view, next_queue, complete_paths)
        return len(candidates)

    def _route_extended(
        self,
        candidate_fh: FileHandle,
        extended: PartialPath,
        candidate_view: StackGraphView,
        next_queue: list[tuple[FileHandle, PartialPath]],
        complete_paths: list[tuple[FileHandle, PartialPath]],
    ) -> None:
        if not _is_complete_path(candidate_view, extended):
            next_queue.append((candidate_fh, extended))
            return
        complete_paths.append((candidate_fh, extended))
        self._stats.complete_paths_found += 1
        if is_import_binding_endpoint(candidate_view, extended.end_node_index):
            next_queue.append((candidate_fh, extended))


def _collect_load_requests(
    paths: list[tuple[FileHandle, PartialPath]],
) -> ForwardStitcherLoadRequests:
    node_requests: list[tuple[FileHandle, int]] = []
    root_requests: list[tuple[FileHandle, PartialSymbolStack]] = []
    seen_nodes: set[tuple[FileHandle, int]] = set()
    seen_roots: set[tuple[FileHandle, PartialSymbolStack]] = set()

    for current_fh, path in paths:
        if path.end_node_index == 0:
            root_key = (current_fh, path.symbol_stack_postcondition)
            if root_key in seen_roots:
                continue
            seen_roots.add(root_key)
            root_requests.append(root_key)
        else:
            node_key = (current_fh, path.end_node_index)
            if node_key in seen_nodes:
                continue
            seen_nodes.add(node_key)
            node_requests.append(node_key)

    return ForwardStitcherLoadRequests(
        node_requests=tuple(node_requests),
        root_requests=tuple(root_requests),
    )


def _is_complete_path(
    def_view: StackGraphView,
    path: PartialPath,
) -> bool:
    if not is_definition_endpoint(def_view, path.end_node_index):
        return False
    start_state = _stack_state_from_partial(
        path.symbol_stack_precondition, path.scope_stack_precondition
    )
    end_state = _stack_state_from_partial(
        path.symbol_stack_postcondition, path.scope_stack_postcondition
    )
    if end_state.scope_stack is not None and path.scope_stack_postcondition.has_variable():
        end_state = StackState(
            symbol_stack=end_state.symbol_stack,
            scope_stack=None,
        )
    if start_state.symbol_stack is not None:
        return False
    if start_state.scope_stack is not None:
        return False
    if end_state.symbol_stack is not None:
        return False
    return end_state.scope_stack is None



import pytest

from blends.models import Graph
from blends.query import (
    adj,
    adj_ast,
    adj_cfg,
    adj_lazy,
    filter_nodes,
    get_node_by_path,
    get_nodes_by_path,
    has_labels,
    match_ast_d,
    match_ast_group,
    match_ast_group_d,
    matching_nodes,
    nodes_by_type,
    pred,
    pred_lazy,
    predicate_has_labels,
)


@pytest.mark.blends_test_group("query_unittesting")
def test_has_labels() -> None:
    n_attrs = {"label_type": "function", "label_name": "test", "label_ast": "AST"}
    assert has_labels(n_attrs, label_type="function")
    assert has_labels(n_attrs, label_type="function", label_name="test")
    assert has_labels(n_attrs, label_type="function", label_name="test", label_ast="AST")
    assert not has_labels(n_attrs, label_type="class")
    assert not has_labels(n_attrs, label_type="function", label_name="other")
    assert has_labels({}, label_type="function") is False
    assert has_labels(n_attrs) is True


@pytest.mark.blends_test_group("query_unittesting")
def test_predicate_has_labels() -> None:
    predicate = predicate_has_labels(label_type="function", label_name="test")
    assert predicate({"label_type": "function", "label_name": "test"})
    assert not predicate({"label_type": "class", "label_name": "test"})
    assert not predicate({"label_type": "function", "label_name": "other"})
    assert not predicate({})


@pytest.mark.blends_test_group("query_unittesting")
def test_filter_nodes() -> None:
    graph = Graph()
    graph.add_node("1", label_type="function")
    graph.add_node("2", label_type="class")
    graph.add_node("3", label_type="function")
    graph.add_node("4", label_type="variable")

    predicate = predicate_has_labels(label_type="function")
    result = filter_nodes(graph, ["1", "2", "3", "4"], predicate)
    assert result == ("1", "3")

    predicate2 = predicate_has_labels(label_type="class")
    result2 = filter_nodes(graph, ["1", "2", "3", "4"], predicate2)
    assert result2 == ("2",)

    result3 = filter_nodes(graph, [], predicate)
    assert result3 == ()


@pytest.mark.blends_test_group("query_unittesting")
def test_matching_nodes() -> None:
    graph = Graph()
    graph.add_node("1", label_type="function", label_name="test")
    graph.add_node("2", label_type="class", label_name="Test")
    graph.add_node("3", label_type="function", label_name="other")
    graph.add_node("4", label_type="function", label_name="test")

    result = matching_nodes(graph, label_type="function")
    assert set(result) == {"1", "3", "4"}

    result2 = matching_nodes(graph, label_type="function", label_name="test")
    assert set(result2) == {"1", "4"}

    result3 = matching_nodes(graph, label_type="class")
    assert result3 == ("2",)

    result4 = matching_nodes(graph, label_type="nonexistent")
    assert result4 == ()


@pytest.mark.blends_test_group("query_unittesting")
def test_adj_and_pred_comprehensive() -> None:  # noqa: PLR0915
    graph = Graph()
    graph.add_node("1")
    graph.add_node("2")
    graph.add_node("3")
    graph.add_node("4")
    graph.add_node("5")
    graph.add_node("6")
    graph.add_edge("1", "2", attr_0="b")
    graph.add_edge("1", "3", attr_0="a")
    graph.add_edge("2", "6", attr_0="b")
    graph.add_edge("3", "4", attr_0="b")
    graph.add_edge("4", "5", attr_1="a")

    assert adj(graph, "1", depth=-1) == ("2", "3", "6", "4", "5")
    assert adj(graph, "1", depth=-1, attr_0="a") == ("3",)
    assert adj(graph, "1", depth=-1, attr_0="b") == ("2", "6")
    assert not adj(graph, "1", depth=0)
    assert adj(graph, "1", depth=1) == ("2", "3")
    assert adj(graph, "1", depth=1, attr_0="a") == ("3",)
    assert adj(graph, "1", depth=1, attr_0="b") == ("2",)
    assert not adj(graph, "1", depth=1, attr_1="b")

    assert pred(graph, "5", depth=-1) == ("4", "3", "1")
    assert not pred(graph, "5", depth=-1, attr_0="a")
    assert pred(graph, "5", depth=-1, attr_1="a") == ("4",)
    assert not pred(graph, "5", depth=-1, attr_0="b")
    assert not pred(graph, "5", depth=0)
    assert pred(graph, "5", depth=1) == ("4",)
    assert not pred(graph, "5", depth=1, attr_0="a")
    assert pred(graph, "5", depth=1, attr_1="a") == ("4",)
    assert not pred(graph, "5", depth=1, attr_0="b")
    assert pred(graph, "5", depth=2) == ("4", "3")


@pytest.mark.blends_test_group("query_unittesting")
def test_adj_cycles() -> None:
    graph = Graph()
    graph.add_node("1")
    graph.add_node("2")
    graph.add_edge("1", "2")
    graph.add_edge("2", "1")

    assert adj(graph, "1", depth=-1) == ("2", "1")
    assert adj(graph, "2", depth=-1) == ("1", "2")
    assert adj(graph, "1", depth=1) == ("2",)
    assert adj(graph, "2", depth=1) == ("1",)


@pytest.mark.blends_test_group("query_unittesting")
def test_adj_ast() -> None:
    graph = Graph()
    graph.add_node("1", label_type="root")
    graph.add_node("2", label_type="function")
    graph.add_node("3", label_type="class")
    graph.add_node("4", label_type="statement")
    graph.add_node("5", label_type="expression")
    graph.add_node("6", label_type="block")

    graph.add_edge("1", "2", label_ast="AST", label_index="0")
    graph.add_edge("1", "3", label_ast="AST", label_index="1")
    graph.add_edge("2", "3", label_ast="AST", label_index="0")

    assert adj_ast(graph, "1") == ("2", "3")
    assert adj_ast(graph, "1", strict=True) == ("2", "3")
    assert adj_ast(graph, "1", label_type="function") == ("2",)
    assert adj_ast(graph, "1", label_type="class") == ("3",)


@pytest.mark.blends_test_group("query_unittesting")
def test_adj_lazy() -> None:
    graph = Graph()
    graph.add_node("1")
    graph.add_node("2")
    graph.add_node("3")
    graph.add_edge("1", "2", attr="a")
    graph.add_edge("1", "3", attr="b")
    graph.add_edge("2", "3", attr="a")

    result = list(adj_lazy(graph, "1", depth=1))
    assert result == ["2", "3"]

    result2 = list(adj_lazy(graph, "1", depth=-1))
    assert result2 == ["2", "3", "3"]

    result3 = list(adj_lazy(graph, "1", depth=1, attr="a"))
    assert result3 == ["2"]

    result4 = list(adj_lazy(graph, "1", depth=0))
    assert result4 == []


@pytest.mark.blends_test_group("query_unittesting")
def test_pred_lazy() -> None:
    graph = Graph()
    graph.add_node("1")
    graph.add_node("2")
    graph.add_node("3")
    graph.add_edge("1", "2", attr="a")
    graph.add_edge("2", "3", attr="b")
    graph.add_edge("1", "3", attr="a")

    result = list(pred_lazy(graph, "3", depth=1))
    assert set(result) == {"1", "2"}

    result2 = list(pred_lazy(graph, "3", depth=-1))
    assert set(result2) == {"1", "2"}

    result3 = list(pred_lazy(graph, "3", depth=1, attr="a"))
    assert result3 == ["1"]

    result4 = list(pred_lazy(graph, "3", depth=0))
    assert result4 == []


@pytest.mark.blends_test_group("query_unittesting")
def test_match_ast_d() -> None:
    graph = Graph()
    graph.add_node("1", label_type="root")
    graph.add_node("2", label_type="function")
    graph.add_node("3", label_type="class")

    graph.add_edge("1", "2", label_ast="AST")
    graph.add_edge("1", "3", label_ast="AST")

    assert match_ast_d(graph, "1", "function", depth=1) == "2"
    assert match_ast_d(graph, "1", "class", depth=1) == "3"
    assert match_ast_d(graph, "1", "nonexistent", depth=1) is None
    assert match_ast_d(graph, "1", "function", depth=2) == "2"


@pytest.mark.blends_test_group("query_unittesting")
def test_match_ast_group() -> None:
    graph = Graph()
    graph.add_node("1", label_type="root")
    graph.add_node("2", label_type="function")
    graph.add_node("3", label_type="function")
    graph.add_node("4", label_type="class")
    graph.add_node("5", label_type="variable")

    graph.add_edge("1", "2", label_ast="AST")
    graph.add_edge("1", "3", label_ast="AST")
    graph.add_edge("1", "4", label_ast="AST")
    graph.add_edge("1", "5", label_ast="AST")

    result = match_ast_group(graph, "1", "function", "class", depth=1)
    assert set(result["function"]) == {"2", "3"}
    assert result["class"] == ["4"]
    assert result["__0__"] == ["5"]

    result2 = match_ast_group(graph, "1", "function", depth=1)
    assert set(result2["function"]) == {"2", "3"}

    result3 = match_ast_group(graph, "1", "nonexistent", depth=1)
    assert result3["nonexistent"] == []


@pytest.mark.blends_test_group("query_unittesting")
def test_match_ast_group_d() -> None:
    graph = Graph()
    graph.add_node("1", label_type="root")
    graph.add_node("2", label_type="function")
    graph.add_node("3", label_type="function")

    graph.add_edge("1", "2", label_ast="AST")
    graph.add_edge("1", "3", label_ast="AST")

    result = match_ast_group_d(graph, "1", "function", depth=1)
    assert set(result) == {"2", "3"}

    result2 = match_ast_group_d(graph, "1", "nonexistent", depth=1)
    assert result2 == []


@pytest.mark.blends_test_group("query_unittesting")
def test_get_nodes_by_path() -> None:
    graph = Graph()
    graph.add_node("1", label_type="root")
    graph.add_node("2", label_type="function")
    graph.add_node("3", label_type="class")
    graph.add_node("4", label_type="method")
    graph.add_node("5", label_type="method")

    graph.add_edge("1", "2", label_ast="AST")
    graph.add_edge("2", "3", label_ast="AST")
    graph.add_edge("3", "4", label_ast="AST")
    graph.add_edge("3", "5", label_ast="AST")

    result = get_nodes_by_path(graph, "1", "function", "class", "method")
    assert set(result) == {"4", "5"}

    result2 = get_nodes_by_path(graph, "1", "function")
    assert result2 == {"2"}

    result3 = get_nodes_by_path(graph, "1", "nonexistent")
    assert result3 == set()

    result4 = get_nodes_by_path(graph, "1")
    assert result4 == set()


@pytest.mark.blends_test_group("query_unittesting")
def test_get_node_by_path() -> None:
    graph = Graph()
    graph.add_node("1", label_type="root")
    graph.add_node("2", label_type="function")
    graph.add_node("3", label_type="class")
    graph.add_node("4", label_type="method")

    graph.add_edge("1", "2", label_ast="AST")
    graph.add_edge("2", "3", label_ast="AST")
    graph.add_edge("3", "4", label_ast="AST")

    assert get_node_by_path(graph, "1", "function", "class", "method") == "4"
    assert get_node_by_path(graph, "1", "function") == "2"
    assert get_node_by_path(graph, "1", "nonexistent") is None
    assert get_node_by_path(graph, "1") is None


@pytest.mark.blends_test_group("query_unittesting")
def test_nodes_by_type() -> None:
    graph = Graph()
    graph.add_node("1", label_type="function")
    graph.add_node("2", label_type="class")
    graph.add_node("3", label_type="function")
    graph.add_node("4", label_type="variable")

    result = nodes_by_type(graph)
    assert set(result["function"]) == {"1", "3"}
    assert result["class"] == ["2"]
    assert result["variable"] == ["4"]

    graph2 = Graph()
    graph2.add_node("1")
    result2 = nodes_by_type(graph2)
    assert result2 == {}


@pytest.mark.blends_test_group("query_unittesting")
def test_nodes_by_type_index_built_at_add_node_time() -> None:
    graph = Graph()
    assert graph.nodes_by_type == {}

    graph.add_node("1", label_type="MethodInvocation")
    assert graph.nodes_by_type == {"MethodInvocation": ["1"]}

    graph.add_node("2", label_type="MethodInvocation")
    assert graph.nodes_by_type["MethodInvocation"] == ["1", "2"]

    graph.add_node("3")
    assert "3" not in graph.nodes_by_type.get("MethodInvocation", [])
    assert len(graph.nodes_by_type) == 1


@pytest.mark.blends_test_group("query_unittesting")
def test_matching_nodes_uses_index() -> None:
    graph = Graph()
    graph.add_node("1", label_type="MethodInvocation", expression="foo")
    graph.add_node("2", label_type="MethodInvocation", expression="bar")
    graph.add_node("3", label_type="VariableDeclaration", expression="foo")

    assert set(matching_nodes(graph, label_type="MethodInvocation")) == {"1", "2"}
    assert matching_nodes(graph, label_type="MethodInvocation", expression="foo") == ("1",)
    assert matching_nodes(graph, label_type="VariableDeclaration") == ("3",)
    assert matching_nodes(graph, label_type="Import") == ()


@pytest.mark.blends_test_group("query_unittesting")
def test_adj_cfg() -> None:
    graph = Graph()
    graph.add_edge("1", "2", label_ast="AST", label_index="0")
    graph.add_edge("1", "3", label_ast="AST", label_index="1")
    graph.add_edge("2", "3", label_ast="AST", label_index="0")
    graph.add_edge("3", "5", label_cfg="CFG", label_index="1")
    graph.add_edge("5", "6", label_cfg="CFG", label_index="1")
    graph.add_edge("1", "4", label_cfg="CFG", label_ast="AST", label_index="2")

    assert adj_cfg(graph, "1") == ("4",)
    assert adj_cfg(graph, "3", strict=True, depth=-1) == ("5", "6")



from collections.abc import Sequence
from typing_extensions import NamedTuple

class Point(NamedTuple):
    row: int
    column: int

class Node:
    type: str
    start_byte: int
    end_byte: int
    start_point: Point
    end_point: Point
    has_error: bool
    children: Sequence[Node]

    def child_by_field_name(self, field_name: str) -> Node | None: ...
    def __hash__(self) -> int: ...

class Tree:
    root_node: Node

    def __init__(self) -> None: ...

class Language:
    def __init__(self, language_ptr: object) -> None: ...

class Parser:
    def __init__(self, language: Language | None = None) -> None: ...
    def parse(
        self, source: bytes, old_tree: Tree | None = None, encoding: str | None = None
    ) -> Tree: ...



from blends.models import (
    NId,
)
from blends.syntax.builders.class_decl import (
    build_class_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]
    name_n_id = n_attrs["label_field_name"]
    name = graph.nodes[name_n_id].get("label_text", "")
    block_id = n_attrs["label_field_body"]
    return build_class_node(args, name, block_id, None)



from collections.abc import (
    Iterator,
)

from blends.symbolic_evaluation.context.search.model import (
    SearchArgs,
    SearchResult,
)


def search(args: SearchArgs) -> Iterator[SearchResult]:
    if (c_id := args.graph.nodes[args.n_id].get("catch_declaration")) and args.graph.nodes[
        c_id
    ].get("variable") == args.symbol:
        yield True, c_id



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.method_declaration import (
    DirectChildren,
    build_method_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    childs = adj_ast(graph, args.n_id)
    accesor_name = graph.nodes[childs[0]]["label_type"]
    block = (
        childs[1]
        if graph.nodes[childs[1]]["label_type"] in {"block", "arrow_expression_clause"}
        else None
    )

    direct: DirectChildren = {}
    if block:
        direct["block_id"] = block

    return build_method_declaration_node(args, accesor_name, direct, {})



from blends.models import (
    NId,
)
from blends.query import (
    match_ast_d,
)
from blends.syntax.builders.import_global import (
    build_import_global_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    expression = ""
    if identifier_n_id := match_ast_d(graph, args.n_id, "identifier"):
        expression = node_to_str(graph, identifier_n_id)
    return build_import_global_node(args, expression, set(), None)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.file import (
    build_file_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    c_ids = adj_ast(graph, args.n_id)

    filtered_ids: list[NId] = []
    for _id in c_ids:
        if graph.nodes[_id]["label_type"] == "statement":
            filtered_ids.extend(adj_ast(args.ast_graph, _id))
            continue
        filtered_ids.append(_id)
    return build_file_node(args, iter(filtered_ids))



from blends.symbolic_evaluation.models import (
    SymbolicEvalArgs,
    SymbolicEvaluation,
)


def evaluate(args: SymbolicEvalArgs) -> SymbolicEvaluation:
    args.evaluation[args.n_id] = False
    return SymbolicEvaluation(args.evaluation[args.n_id], args.triggers)



from blends.models import (
    NId,
)
from blends.query import (
    match_ast,
)
from blends.syntax.builders.method_declaration import (
    DirectChildren,
    build_method_declaration_node,
)
from blends.syntax.metadata.java import (
    add_method_to_metadata,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.syntax.readers.java.class_declaration import (
    extract_modifiers,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]
    name_id = n_attrs["label_field_name"]
    name = node_to_str(graph, name_id)

    block_id = n_attrs.get("label_field_body")

    parameters_id = n_attrs["label_field_parameters"]
    if "__0__" not in match_ast(graph, parameters_id, "(", ")"):
        parameters_id = None

    modifiers_id, access_modifiers = extract_modifiers(graph, args.n_id)

    direct: DirectChildren = {}
    if block_id:
        direct["block_id"] = block_id
    if modifiers_id:
        direct["modifiers_id"] = modifiers_id
    if parameters_id:
        direct["parameters_id"] = parameters_id

    if args.syntax_graph.nodes.get(NId(0)):
        add_method_to_metadata(args, name)

    return build_method_declaration_node(args, name, direct, {}, access_modifiers)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.cfg.dispatchers.multifile import (
    adjust_return_value,
)
from blends.syntax.models import (
    SyntaxCfgArgs,
)

DEC_TYPES = {
    "Class",
    "MethodDeclaration",
    "VariableDeclaration",
}


def build(args: SyntaxCfgArgs) -> NId:
    if (
        args.is_multifile
        and args.graph.nodes[args.n_id]["label_type"] == "Return"
        and args.graph.nodes[args.n_id].get("value_id")
    ):
        adjust_return_value(args)

    for _id in adj_ast(args.graph, args.n_id):
        if args.graph.nodes[_id]["label_type"] not in DEC_TYPES:
            continue

        args.graph.add_edge(
            args.n_id,
            args.generic(args.fork(_id, args.nxt_id)),
            label_cfg="CFG",
        )

    return args.n_id



from blends.ctx import ctx
from blends.models import NId
from blends.query import adj_ast
from blends.syntax.builders.parameter import build_parameter_node
from blends.syntax.builders.parameter_list import build_parameter_list_node
from blends.syntax.models import SyntaxGraphArgs

_IGNORED_PARAMETER_TYPES = {",", "keyword_separator"}


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    _, *c_ids, _ = adj_ast(graph, args.n_id)

    if ctx.has_feature_flag("StackGraph"):
        return _build_stack_graph_parameter_list(args, c_ids)

    return build_parameter_list_node(
        args,
        c_ids=(
            _id for _id in c_ids if graph.nodes[_id]["label_type"] not in _IGNORED_PARAMETER_TYPES
        ),
    )


def _build_stack_graph_parameter_list(args: SyntaxGraphArgs, c_ids: list[NId]) -> NId:
    graph = args.ast_graph
    args.syntax_graph.add_node(args.n_id, label_type="ParameterList")

    for c_id in c_ids:
        label_type = graph.nodes[c_id]["label_type"]
        if label_type in _IGNORED_PARAMETER_TYPES:
            continue

        if label_type == "identifier":
            child_id = build_parameter_node(
                args=args.fork_n_id(c_id),
                variable=graph.nodes[c_id]["label_text"],
                variable_type=None,
                value_id=None,
            )
        else:
            child_id = args.generic(args.fork_n_id(c_id))

        args.syntax_graph.add_edge(args.n_id, child_id, label_ast="AST")

    return args.n_id



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_d,
)
from blends.syntax.builders.method_declaration import (
    DirectChildren,
    build_method_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]
    body_id = n_attrs["label_field_body"]
    param_id = (
        match_ast_d(graph, args.n_id, "identifier")
        or match_ast_d(graph, args.n_id, "parameter_list")
        or adj_ast(graph, args.n_id)[0]
    )

    direct: DirectChildren = {"block_id": body_id, "parameters_id": param_id}

    return build_method_declaration_node(args, "LambdaExpression", direct, {})



from blends.symbolic_evaluation.context.search import definition_search, search
from blends.symbolic_evaluation.evaluate import (
    evaluate,
    get_node_evaluation_results,
    get_node_evaluation_results_advanced,
)
from blends.symbolic_evaluation.utils import (
    get_backward_paths,
    get_forward_paths,
    get_object_identifiers,
)

__all__ = [
    "definition_search",
    "evaluate",
    "get_backward_paths",
    "get_forward_paths",
    "get_node_evaluation_results",
    "get_node_evaluation_results_advanced",
    "get_object_identifiers",
    "search",
]



from blends.models import (
    Graph,
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.file import (
    build_file_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def get_filtered_node(ast_graph: Graph, c_id: NId) -> NId:
    if ast_graph.nodes[c_id]["label_type"] == "expression_statement":
        return adj_ast(ast_graph, c_id)[0]
    return c_id


def reader(args: SyntaxGraphArgs) -> NId:
    c_ids = [
        c_id
        for n_id in adj_ast(args.ast_graph, args.n_id)
        if args.ast_graph.nodes[n_id].get("label_type") != "empty_statement"
        and (c_id := get_filtered_node(args.ast_graph, n_id))
    ]

    return build_file_node(args, c_ids)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.binary_operation import (
    build_binary_operation_node,
)
from blends.syntax.builders.expression_statement import (
    build_expression_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    c_ids = adj_ast(graph, args.n_id)
    invalid_childs = {";", "type_test", "(", ")"}
    expected_children_for_binary_operation = 3

    if len(c_ids) == expected_children_for_binary_operation:
        operator = node_to_str(graph, c_ids[1])
        return build_binary_operation_node(args, operator, c_ids[0], c_ids[2])

    return build_expression_statement_node(
        args,
        c_ids=(_id for _id in c_ids if graph.nodes[_id]["label_type"] not in invalid_childs),
    )



from blends.models import (
    NId,
)
from blends.syntax.builders.reserved_word import (
    build_reserved_word_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    val_str = args.ast_graph.nodes[args.n_id].get("label_text") or node_to_str(
        args.ast_graph,
        args.n_id,
    )
    return build_reserved_word_node(args, f"OperatorValue {val_str}")



import pytest

from blends.models import SyntaxGraph
from blends.stack.attributes import (
    STACK_GRAPH_IS_DEFINITION,
    STACK_GRAPH_IS_REFERENCE,
    STACK_GRAPH_KIND,
    STACK_GRAPH_SYMBOL,
)
from blends.stack.node_kinds import StackGraphNodeKind
from blends.stack.partial_path import (
    PartialPathLimits,
    compute_minimal_partial_paths_in_file,
)
from blends.stack.partial_path_database import (
    PartialPathDatabase,
    StackGraphViewMetadata,
    remap_symbol_id,
)
from blends.stack.view import StackGraphView, SymbolInterner


def _build_reference_definition_graph() -> SyntaxGraph:
    graph = SyntaxGraph()
    graph.add_node("1")
    graph.update_node_attribute("1", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SYMBOL.value)
    graph.update_node_attribute("1", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("1", STACK_GRAPH_IS_REFERENCE, value=True)

    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SYMBOL.value)
    graph.update_node_attribute("2", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("2", STACK_GRAPH_IS_DEFINITION, value=True)

    graph.add_edge("1", "2", precedence=0)
    return graph


@pytest.mark.blends_test_group("stack_unittesting")
def test_partial_path_database_roundtrip_and_copies_stats() -> None:
    graph = _build_reference_definition_graph()
    view = StackGraphView.from_syntax_graph(graph, path="test.py")
    per_file = compute_minimal_partial_paths_in_file(
        view,
        "test.py",
        limits=PartialPathLimits(),
    )

    database = PartialPathDatabase()
    database.put_file(file_handle=1, file_path="test.py", view=view, partial_paths=per_file)

    record = database.get_file(1)
    assert record is not None
    assert record.file_handle == 1
    assert record.file_path == "test.py"
    assert record.view_snapshot.symbol_id_to_string == ("x",)
    assert record.paths == per_file.paths

    record.stats.mark_limit_hit("external")
    changed_record = database.get_file(1)
    assert changed_record is not None
    assert "external" not in changed_record.stats.limit_hits

    assert database.list_files() == (1,)
    database.drop_file(1)
    assert database.get_file(1) is None


@pytest.mark.blends_test_group("stack_unittesting")
def test_remap_symbol_id_interns_into_target_interner() -> None:
    metadata = StackGraphViewMetadata(file_handles=("test.py",), symbol_id_to_string=("x",))
    target = SymbolInterner(symbol_to_id={}, id_to_symbol=[])

    symbol_id = remap_symbol_id(0, metadata, target)

    assert symbol_id == 0
    assert target.lookup(symbol_id) == "x"



from blends.models import (
    NId,
)
from blends.syntax.builders.namespace import (
    build_namespace_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]

    name_id = n_attrs["label_field_name"]
    name_str = node_to_str(graph, name_id)
    block_id = n_attrs.get("label_field_body")

    return build_namespace_node(args, name_str, block_id)



from blends.models import (
    NId,
)
from blends.syntax.builders.parameter import (
    build_parameter_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    var_name_id = graph.nodes[args.n_id].get("label_field_name")
    variable = node_to_str(graph, var_name_id) if var_name_id else node_to_str(graph, args.n_id)

    var_type_id = graph.nodes[args.n_id].get("label_field_type")
    var_type = node_to_str(graph, var_type_id) if var_type_id else None
    return build_parameter_node(args=args, variable=variable, variable_type=var_type, value_id=None)



from blends.models import (
    NId,
)
from blends.syntax.builders.unary_expression import (
    build_unary_expression_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    node_id = args.ast_graph.nodes[args.n_id]
    operator = args.ast_graph.nodes[node_id["label_field_operator"]]["label_text"]
    operand = node_id["label_field_operand"]
    return build_unary_expression_node(args, operator, operand)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_d,
    match_ast_group_d,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph

    body_id = match_ast_d(graph, args.n_id, "switch")
    if not body_id:
        body_id = adj_ast(graph, args.n_id)[0]

    value_id = graph.nodes[args.n_id].get("label_field_initializer") or graph.nodes[args.n_id].get(
        "label_field_value",
    )

    if not value_id:
        value_id = adj_ast(graph, args.n_id)[2]

    # Go parser does not include a SwitchBody node out of the box
    args.syntax_graph.add_node(
        args.n_id,
        block_id=body_id,
        value_id=value_id,
        label_type="SwitchStatement",
    )
    args.syntax_graph.add_edge(
        args.n_id,
        body_id,
        label_ast="AST",
    )
    args.syntax_graph.add_edge(
        args.n_id,
        args.generic(args.fork_n_id(value_id)),
        label_ast="AST",
    )

    case_ids = (
        match_ast_group_d(graph, args.n_id, "expression_case")
        + match_ast_group_d(graph, args.n_id, "default_case")
        + match_ast_group_d(graph, args.n_id, "type_case")
    )

    args.syntax_graph.add_node(
        body_id,
        label_type="SwitchBody",
    )

    for c_id in case_ids:
        args.syntax_graph.add_edge(
            body_id,
            args.generic(args.fork_n_id(c_id)),
            label_ast="AST",
        )

    return args.n_id



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from blends.stack.criteria import (
    is_definition_endpoint,
)

if TYPE_CHECKING:
    from blends.stack.view import (
        StackGraphView,
    )


@dataclass(frozen=True, slots=True)
class PartialPathEdge:
    source_node_index: int
    precedence: int


@dataclass(frozen=True, slots=True)
class DefinitionCandidate:
    definition_node_index: int
    edges: tuple[PartialPathEdge, ...]


def edge_shadows(a: PartialPathEdge, b: PartialPathEdge) -> bool:
    return a.source_node_index == b.source_node_index and a.precedence > b.precedence


def edge_list_shadows(
    a_edges: tuple[PartialPathEdge, ...], b_edges: tuple[PartialPathEdge, ...]
) -> bool:
    if not a_edges or not b_edges:
        return False
    for a_edge, b_edge in zip(a_edges, b_edges, strict=False):
        if a_edge.source_node_index != b_edge.source_node_index:
            return False
        if edge_shadows(a_edge, b_edge):
            return True
    return False


def prune_shadowed_candidates(candidates: list[DefinitionCandidate]) -> list[DefinitionCandidate]:
    kept: list[DefinitionCandidate] = []
    for candidate in sorted(candidates, key=lambda c: c.definition_node_index):
        if any(edge_list_shadows(existing.edges, candidate.edges) for existing in kept):
            continue
        kept = [
            existing for existing in kept if not edge_list_shadows(candidate.edges, existing.edges)
        ]
        kept.append(candidate)
    return kept


def sort_candidates_deterministically(
    candidates: list[DefinitionCandidate],
) -> list[DefinitionCandidate]:
    def sort_key(candidate: DefinitionCandidate) -> tuple[tuple[int, ...], tuple[int, ...], int]:
        edge_sources = tuple(edge.source_node_index for edge in candidate.edges)
        edge_precedences = tuple(-edge.precedence for edge in candidate.edges)
        return (edge_sources, edge_precedences, candidate.definition_node_index)

    return sorted(candidates, key=sort_key)


def select_definition_candidates_from_scope(
    view: StackGraphView, scope_node_index: int, symbol_id: int
) -> list[int]:
    candidates: list[DefinitionCandidate] = []
    for sink_index, edge_precedence in view.outgoing[scope_node_index]:
        if not is_definition_endpoint(view, sink_index):
            continue
        if view.symbol_id_at(sink_index) != symbol_id:
            continue
        candidates.append(
            DefinitionCandidate(
                definition_node_index=sink_index,
                edges=(PartialPathEdge(scope_node_index, edge_precedence),),
            )
        )

    pruned = prune_shadowed_candidates(candidates)
    ordered = sort_candidates_deterministically(pruned)
    return [candidate.definition_node_index for candidate in ordered]



from blends.content.custom_parsers.helm_templates import helm_parser
from blends.models import CustomParser, Language

custom_parsers_by_language: dict[Language, list[CustomParser]] = {
    Language.JSON: [helm_parser],
    Language.YAML: [helm_parser],
}



from __future__ import annotations

from typing import TYPE_CHECKING

from blends.stack.policies.base import ExportPolicy
from blends.stack.scope_types import ScopeType

if TYPE_CHECKING:
    from blends.models import NId, SyntaxGraph


class GoExportPolicy(ExportPolicy):
    def is_scope_exported(
        self,
        _graph: SyntaxGraph,
        _node_id: NId,
        scope_type: str | None,
    ) -> bool:
        return scope_type == ScopeType.FILE.value

    def is_definition_exported(
        self,
        graph: SyntaxGraph,
        node_id: NId,
        parent_scope_id: NId | None,
    ) -> bool:
        if parent_scope_id is not None:
            scope_type = graph.nodes[parent_scope_id].get("stack_graph_scope_type")
            if scope_type != ScopeType.FILE.value:
                return False
        symbol = graph.nodes[node_id].get("stack_graph_symbol")
        if isinstance(symbol, str) and len(symbol) > 0:
            return bool(symbol[0].isupper())
        name = graph.nodes[node_id].get("name")
        if isinstance(name, str) and len(name) > 0:
            return bool(name[0].isupper())
        return False



import pytest

from blends.models import (
    SyntaxGraph,
)
from blends.stack.edges import (
    Edge,
    add_edge,
)


@pytest.mark.blends_test_group("stack_unittesting")
def test_add_edge_from_reference_to_scope() -> None:
    graph = SyntaxGraph()
    graph.add_node("ref_1", label_type="SymbolLookup", symbol="x")
    graph.add_node("scope_1", label_type="File")

    add_edge(graph, Edge(source="ref_1", sink="scope_1", precedence=0))

    assert graph.has_edge("ref_1", "scope_1")
    edge_data = graph["ref_1"]["scope_1"]
    assert edge_data["precedence"] == 0
    assert "label_stack" not in edge_data
    assert "symbol" not in edge_data


@pytest.mark.blends_test_group("stack_unittesting")
def test_add_edge_from_scope_to_definition() -> None:
    graph = SyntaxGraph()
    graph.add_node("scope_1", label_type="File")
    graph.add_node("def_1", label_type="VariableDeclaration", variable="x")

    add_edge(graph, Edge(source="scope_1", sink="def_1", precedence=0))

    assert graph.has_edge("scope_1", "def_1")
    edge_data = graph["scope_1"]["def_1"]
    assert edge_data["precedence"] == 0
    assert "label_stack" not in edge_data
    assert "symbol" not in edge_data


@pytest.mark.blends_test_group("stack_unittesting")
def test_add_lookup_edge() -> None:
    graph = SyntaxGraph()
    graph.add_node("child_scope", label_type="MethodDeclaration")
    graph.add_node("parent_scope", label_type="File")

    add_edge(graph, Edge(source="child_scope", sink="parent_scope", precedence=0))

    assert graph.has_edge("child_scope", "parent_scope")
    edge_data = graph["child_scope"]["parent_scope"]
    assert edge_data["precedence"] == 0
    assert "label_stack" not in edge_data
    assert "symbol" not in edge_data


@pytest.mark.blends_test_group("stack_unittesting")
def test_add_edge_from_reference_to_scope_with_precedence() -> None:
    graph = SyntaxGraph()
    graph.add_node("ref_1", label_type="SymbolLookup", symbol="x")
    graph.add_node("scope_1", label_type="File")

    add_edge(graph, Edge(source="ref_1", sink="scope_1", precedence=5))

    edge_data = graph["ref_1"]["scope_1"]
    assert edge_data["precedence"] == 5  # noqa: PLR2004


@pytest.mark.blends_test_group("stack_unittesting")
def test_add_edge_from_scope_to_definition_with_precedence() -> None:
    graph = SyntaxGraph()
    graph.add_node("scope_1", label_type="MethodDeclaration")
    graph.add_node("def_1", label_type="VariableDeclaration", variable="x")

    add_edge(graph, Edge(source="scope_1", sink="def_1", precedence=10))

    edge_data = graph["scope_1"]["def_1"]
    assert edge_data["precedence"] == 10  # noqa: PLR2004


@pytest.mark.blends_test_group("stack_unittesting")
def test_multiple_edges_from_scope_to_definitions() -> None:
    graph = SyntaxGraph()
    graph.add_node("scope_1", label_type="File")
    graph.add_node("def_1", label_type="VariableDeclaration", variable="x")
    graph.add_node("def_2", label_type="VariableDeclaration", variable="y")

    add_edge(graph, Edge(source="scope_1", sink="def_1", precedence=0))
    add_edge(graph, Edge(source="scope_1", sink="def_2", precedence=0))

    assert graph.has_edge("scope_1", "def_1")
    assert graph.has_edge("scope_1", "def_2")
    assert graph["scope_1"]["def_1"]["precedence"] == 0
    assert graph["scope_1"]["def_2"]["precedence"] == 0


@pytest.mark.blends_test_group("stack_unittesting")
def test_scope_hierarchy_with_lookup_edges() -> None:
    graph = SyntaxGraph()
    graph.add_node("file_scope", label_type="File")
    graph.add_node("class_scope", label_type="Class")
    graph.add_node("method_scope", label_type="MethodDeclaration")

    add_edge(graph, Edge(source="class_scope", sink="file_scope", precedence=0))
    add_edge(graph, Edge(source="method_scope", sink="class_scope", precedence=0))

    assert graph.has_edge("class_scope", "file_scope")
    assert graph.has_edge("method_scope", "class_scope")
    assert graph["class_scope"]["file_scope"]["precedence"] == 0
    assert graph["method_scope"]["class_scope"]["precedence"] == 0



from blends.models import (
    NId,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def build_reserved_word_node(args: SyntaxGraphArgs, value: str) -> NId:
    args.syntax_graph.add_node(
        args.n_id,
        value=value,
        label_type="ReservedWord",
    )

    return args.n_id



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_d,
)
from blends.syntax.builders.file import (
    build_file_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    if (body_id := match_ast_d(args.ast_graph, args.n_id, "document")) and (
        block_id := match_ast_d(args.ast_graph, body_id, "block_node")
    ):
        filtered_ids = (
            _id
            for _id in adj_ast(args.ast_graph, block_id)
            if args.ast_graph.nodes[_id]["label_type"] != "---"
        )
        return build_file_node(args, iter(filtered_ids))

    return args.n_id



from blends.models import (
    NId,
)
from blends.query import (
    match_ast,
)
from blends.syntax.builders.binary_operation import (
    build_binary_operation_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    left_id = args.ast_graph.nodes[args.n_id]["label_field_left"]
    if graph.nodes[left_id]["label_type"] == "parenthesized_expression":
        left_id = match_ast(graph, left_id)["__1__"]
    right_id = args.ast_graph.nodes[args.n_id]["label_field_right"]
    if graph.nodes[right_id]["label_type"] == "parenthesized_expression":
        right_id = match_ast(graph, right_id)["__1__"]
    operator_id = args.ast_graph.nodes[args.n_id]["label_field_operator"]
    operator = args.ast_graph.nodes[operator_id]["label_text"]
    return build_binary_operation_node(args, operator, left_id, right_id)



from collections.abc import (
    Iterator,
)

from blends.ctx import ctx
from blends.models import (
    NId,
    SyntaxGraph,
)
from blends.query import (
    adj_cfg,
    lookup_first_cfg_parent,
    matching_nodes,
    pred_ast,
    pred_cfg,
    search_pred_until_type,
)
from blends.symbolic_evaluation.models import (
    Path,
)


def iter_backward_paths(graph: SyntaxGraph, cfg_n_id: NId) -> Iterator[Path]:
    path = [cfg_n_id]
    parents = pred_cfg(graph, cfg_n_id)

    if not parents:
        yield path

    for parent in parents:
        for sub_path in iter_backward_paths(graph, parent):
            yield path + sub_path


def get_backward_paths(graph: SyntaxGraph, n_id: NId) -> Iterator[Path]:
    cfg_id = lookup_first_cfg_parent(graph, n_id)
    paths_gen = iter_backward_paths(graph, cfg_id)
    if limit := ctx.recursion_limit:
        it_count = 1
        while it_count < limit:
            try:
                yield next(paths_gen)
                it_count += 1
            except StopIteration:
                return
    else:
        yield from paths_gen


def iter_forward_paths(graph: SyntaxGraph, cfg_id: NId) -> Iterator[Path]:
    path = [cfg_id]
    childs = adj_cfg(graph, cfg_id)

    if not childs:
        yield path

    for child in childs:
        for sub_path in iter_forward_paths(graph, child):
            yield path + sub_path


def get_forward_paths(graph: SyntaxGraph, n_id: NId) -> Iterator[Path]:
    cfg_id = lookup_first_cfg_parent(graph, n_id)
    yield from iter_forward_paths(graph, cfg_id)


def get_lookup_path(graph: SyntaxGraph, path: Path, symbol_id: NId) -> Path:
    cfg_parent = lookup_first_cfg_parent(graph, symbol_id)
    cfg_parent_idx = path.index(cfg_parent)  # current instruction idx
    return path[cfg_parent_idx + 1 :]  # from previous instruction idx


def get_object_identifiers(graph: SyntaxGraph, obj_names: set[str]) -> list[str]:
    return [
        var_name
        for n_id in matching_nodes(graph, label_type="ObjectCreation")
        if graph.nodes[n_id].get("name") in obj_names
        and (pred := pred_ast(graph, n_id)[0])
        and graph.nodes[pred].get("label_type") == "VariableDeclaration"
        and (var_name := graph.nodes[pred].get("variable"))
    ]


def get_current_class(graph: SyntaxGraph, n_id: NId) -> NId:
    return class_nids[0] if (class_nids := search_pred_until_type(graph, n_id, {"Class"})) else n_id



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_d,
)
from blends.syntax.builders.switch_section import (
    build_switch_section_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    statements_id = match_ast_d(graph, args.n_id, "statements")
    if not statements_id:
        statements_id = adj_ast(graph, args.n_id)[-1]
    child_ids = adj_ast(graph, statements_id)

    val_id = match_ast_d(graph, args.n_id, "switch_pattern")
    if not val_id:
        val_id = adj_ast(graph, args.n_id)[0]
    case_expr = node_to_str(graph, val_id)

    return build_switch_section_node(args, case_expr, child_ids)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_group_d,
)
from blends.syntax.builders.string_literal import (
    build_string_literal_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = args.ast_graph.nodes[args.n_id]
    label_text = node_to_str(graph, args.n_id)
    if text_nid := n_attrs.get("label_field_text"):
        label_text = args.ast_graph.nodes[text_nid].get("label_text", "")

    ints = match_ast_group_d(graph, args.n_id, "interpolated_expression")
    int_subs = [adj_ast(graph, _id)[0] for _id in ints]

    if len(int_subs) > 0:
        return build_string_literal_node(args, label_text, iter(int_subs))

    return build_string_literal_node(args, label_text)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.variable_declaration import (
    build_variable_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    name_id = args.ast_graph.nodes[args.n_id]["label_field_name"]
    name = node_to_str(args.ast_graph, name_id)
    body_id = args.ast_graph.nodes[args.n_id]["label_field_body"]
    clause = adj_ast(args.ast_graph, args.n_id, label_type="extends_clause")
    var_type = None
    if clause:
        var_type = node_to_str(args.ast_graph, clause[0]).replace("extends", "")
    return build_variable_declaration_node(args, name, var_type, body_id)



C_SHARP_EXPRESSION = {
    "anonymous_method_expres",
    "anonymous_object_creation_expression",
    "array_creation_expression",
    "as_expression",
    "assignment_expression",
    "await_expression",
    "base_expression",
    "binary_expression",
    "boolean_literal",
    "cast_expression",
    "character_literal",
    "checked_expression",
    "conditional_access_expression",
    "conditional_expression",
    "default_expression",
    "element_access_expression",
    "element_binding_expression",
    "generic_name",
    "global",
    "identifier",
    "implicit_array_creation_expression",
    "implicit_object_creation_expression",
    "implicit_stack_alloc_array_creation_expression",
    "initializer_expression",
    "integer_literal",
    "interpolated_string_expression",
    "invocation_expression",
    "is_expression",
    "is_pattern_expression",
    "lambda_expression",
    "make_ref_expression",
    "member_access_expression",
    "null_literal",
    "object_creation_expression",
    "parenthesized_expression",
    "postfix_unary_expression",
    "prefix_unary_expression",
    "query_expression",
    "range_expression",
    "real_literal",
    "ref_expression",
    "ref_type_expression",
    "ref_value_expression",
    "size_of_expression",
    "stack_alloc_array_creation_expression",
    "string_literal",
    "switch_expression",
    "this_expression",
    "throw_expression",
    "tuple_expression",
    "type_of_expression",
    "verbatim_string_literal",
    "with_expression",
}


C_SHARP_STATEMENT = {
    "block",
    "break_statement",
    "checked_statement",
    "continue_statement",
    "do_statement",
    "empty_statement",
    "expression_statement",
    "fixed_statement",
    "for_each_statement",
    "for_statement",
    "goto_statement",
    "if_statement",
    "labeled_statement",
    "local_declaration_statement",
    "local_function_statement",
    "lock_statement",
    "return_statement",
    "switch_statement",
    "throw_statement",
    "try_statement",
    "unsafe_statement",
    "using_statement",
    "while_statement",
    "yield_statement",
}

JAVA_STATEMENT = {
    "declaration",
    "expression_statement",
    "labeled_statement",
    "if_statement",
    "while_statement",
    "for_statement",
    "enhanced_for_statement",
    "block",
    "assert_statement",
    "do_statement",
    "return_statement",
    "yield_statement",
    "switch_expression",
    "synchronized_statement",
    "local_variable_declaration",
    "throw_statement",
    "try_statement",
    "try_with_resources_statement",
}

JAVASCRIPT_STATEMENT = {
    "export_statement",
    "import_statement",
    "debugger_statement",
    "declaration",
    "statement_block",
    "expression_statement",
    "if_statement",
    "while_statement",
    "for_statement",
    "switch_statement",
    "for_in_statement",
    "do_statement",
    "try_statement",
    "with_statement",
    "break_statement",
    "continue_statement",
    "return_statement",
    "throw_statement",
    "empty_statement",
    "labeled_statement",
}

TYPESCRIPT_PRIMARY_TYPES = {
    "array_type",
    "conditional_type",
    "existential_type",
    "flow_maybe_type",
    "generic_type",
    "index_type_query",
    "intersection_type",
    "literal_type",
    "lookup_type",
    "nested_type_identifier",
    "object_type",
    "parenthesized_type",
    "predefined_type",
    "template_literal_type",
    "this_type",
    "tuple_type",
    "type_identifier",
    "type_query",
    "union_type",
}



from collections.abc import (
    Iterable,
)

from blends.ctx import ctx
from blends.models import (
    NId,
)
from blends.stack.edges import (
    Edge,
    add_edge,
)
from blends.stack.node_helpers import (
    pop_scoped_symbol_node_attributes,
    scope_node_attributes,
)
from blends.stack.policies.registry import (
    get_export_policy,
)
from blends.stack.scope_types import ScopeType
from blends.syntax.builders.utils import (
    get_next_binding_precedence,
    get_next_synthetic_node_id,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def _enter_scope_stack(args: SyntaxGraphArgs, class_name: str) -> None:
    if not ctx.has_feature_flag("StackGraph"):
        return
    policy = get_export_policy(args.language)
    args.syntax_graph.update_node(
        args.n_id,
        scope_node_attributes(
            is_exported=policy.is_scope_exported(
                args.syntax_graph, args.n_id, ScopeType.CLASS.value
            ),
            scope_type=ScopeType.CLASS,
        ),
    )
    scope_stack = args.metadata.setdefault("scope_stack", [])
    if scope_stack:
        add_edge(
            args.syntax_graph,
            Edge(source=args.n_id, sink=scope_stack[-1], precedence=0),
        )
    scope_stack.append(args.n_id)
    class_member_scope_stack = args.metadata.setdefault("class_member_scope_stack", [])
    class_member_scope_stack.append((args.n_id, class_name))


def _exit_scope_stack_if_current(args: SyntaxGraphArgs) -> None:
    if not ctx.has_feature_flag("StackGraph"):
        return
    scope_stack = args.metadata.setdefault("scope_stack", [])
    if scope_stack and scope_stack[-1] == args.n_id:
        scope_stack.pop()
    class_member_scope_stack = args.metadata.setdefault("class_member_scope_stack", [])
    if class_member_scope_stack and class_member_scope_stack[-1][0] == args.n_id:
        class_member_scope_stack.pop()


def _setup_stack_graph_definition(args: SyntaxGraphArgs, name: str) -> None:
    scope_stack = args.metadata.setdefault("scope_stack", [])
    if not scope_stack:
        return
    parent_scope = scope_stack[-1]
    precedence = get_next_binding_precedence(args)
    class_def_id = get_next_synthetic_node_id(args)
    args.syntax_graph.add_node(
        class_def_id,
        label_type="Class",
        name=name,
        sg_node_type="definition",
    )
    args.syntax_graph.update_node(
        class_def_id,
        pop_scoped_symbol_node_attributes(
            symbol=name,
            precedence=precedence,
        ),
    )
    add_edge(
        args.syntax_graph,
        Edge(source=parent_scope, sink=class_def_id, precedence=precedence),
    )
    add_edge(
        args.syntax_graph,
        Edge(source=class_def_id, sink=args.n_id, precedence=0),
    )
    ast_node = args.ast_graph.nodes.get(args.n_id, {})
    if (label_l := ast_node.get("label_l")) is not None:
        args.syntax_graph.update_node_attribute(class_def_id, "label_l", label_l)
        args.syntax_graph.update_node_attribute(
            class_def_id, "label_c", ast_node.get("label_c", "0")
        )


def build_class_node(  # noqa: PLR0913
    args: SyntaxGraphArgs,
    name: str,
    block_id: NId | None,
    attrl_ids: Iterable[NId] | None,
    inherited_class: str | None = None,
    modifiers_id: NId | None = None,
    access_modifiers: str | None = None,
) -> NId:
    args.syntax_graph.add_node(
        args.n_id,
        name=name,
        block_id=block_id,
        label_type="Class",
        **({"access_modifiers": access_modifiers} if access_modifiers else {}),
    )

    if ctx.has_feature_flag("StackGraph"):
        _setup_stack_graph_definition(args, name)

    _enter_scope_stack(args, name)

    if block_id:
        args.syntax_graph.add_edge(
            args.n_id,
            args.generic(args.fork_n_id(block_id)),
            label_ast="AST",
        )

    if attrl_ids:
        for n_id in attrl_ids:
            args.syntax_graph.add_edge(
                args.n_id,
                args.generic(args.fork_n_id(n_id)),
                label_ast="AST",
            )

    if inherited_class:
        args.syntax_graph.update_node_attribute(args.n_id, "inherited_class", inherited_class)

    if modifiers_id:
        args.syntax_graph.update_node_attribute(args.n_id, "modifiers_id", modifiers_id)
        args.syntax_graph.add_edge(
            args.n_id,
            args.generic(args.fork_n_id(modifiers_id)),
            label_ast="AST",
        )

    if args.metadata["class_path"]:
        args.metadata["class_path"].pop()

    _exit_scope_stack_if_current(args)

    return args.n_id



import pytest

from test.stack.test_helpers import (
    count_defs_refs,
    resolve_symbol,
    setup_stack_resolution,
)


@pytest.mark.blends_test_group("functional")
def test_multiple_bare_params() -> None:
    view, record = setup_stack_resolution("case_multiple_params.py")
    for symbol in ["a", "b", "c"]:
        defs, refs = count_defs_refs(view, symbol)
        result = resolve_symbol(view, record, symbol)
        assert defs == 1, f"Expected 1 def of '{symbol}', got {defs}"
        assert refs == 1, f"Expected 1 ref to '{symbol}', got {refs}"
        assert len(result) == 1, f"Expected bare param '{symbol}' to resolve, got {len(result)}"


@pytest.mark.blends_test_group("functional")
def test_self_reference_in_method() -> None:
    view, record = setup_stack_resolution("case_self_reference.py")
    defs, refs = count_defs_refs(view, "self")
    result = resolve_symbol(view, record, "self")
    assert defs == 1, f"Expected 1 def of 'self', got {defs}"
    assert refs >= 1, f"Expected at least 1 ref to 'self', got {refs}"
    assert len(result) == 1, (
        f"Expected 'self' parameter to resolve to its definition, got {len(result)}"
    )



from blends.models import (
    NId,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def build_literal_node(args: SyntaxGraphArgs, value: str, var_type: str) -> NId:
    args.syntax_graph.add_node(
        args.n_id,
        value=value,
        value_type=var_type,
        label_type="Literal",
    )

    return args.n_id



from blends.models import (
    NId,
)
from blends.query import (
    get_node_by_path,
    match_ast_d,
    match_ast_group_d,
)
from blends.syntax.builders.jsx_element import (
    build_jsx_element_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    childs_with_attrs = match_ast_group_d(graph, args.n_id, "jsx_attribute")
    childs = match_ast_group_d(graph, args.n_id, "jsx_opening_element") + match_ast_group_d(
        graph,
        args.n_id,
        "jsx_self_closing_element",
    )
    for _id in childs:
        jsx_attributes = match_ast_group_d(graph, _id, "jsx_attribute")
        identifier = match_ast_d(graph, _id, "identifier")
        nested_identifier = match_ast_d(graph, _id, "nested_identifier")
        spread_element = get_node_by_path(graph, _id, "jsx_expression", "spread_element")
        if jsx_attributes and len(childs) == 1:
            childs_with_attrs.extend(jsx_attributes)
        elif jsx_attributes:
            childs_with_attrs.append(_id)
        if identifier:
            childs_with_attrs.append(identifier)
        if nested_identifier:
            childs_with_attrs.append(nested_identifier)
        if spread_element:
            childs_with_attrs.append(spread_element)

    if jsx_elements_nested := match_ast_group_d(graph, args.n_id, "jsx_element"):
        childs_with_attrs.extend(jsx_elements_nested)

    return build_jsx_element_node(args, childs_with_attrs)



from blends.models import (
    NId,
)
from blends.query import (
    match_ast,
)
from blends.syntax.builders.variable_declaration import (
    build_variable_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    identifier_id = match_ast(
        graph,
        args.n_id,
        "identifier",
    )
    ident_id = identifier_id.get("identifier")
    var_name = graph.nodes[ident_id]["label_text"] if ident_id else ""
    return build_variable_declaration_node(args, var_name, None, ident_id)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.element_access import (
    build_element_access_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_id = args.n_id

    expr_id = graph.nodes[n_id]["label_field_object"]

    labels_to_ignore = {"[", "]"}
    c_ids = [
        node_id
        for node_id in adj_ast(graph, n_id)
        if graph.nodes[node_id]["label_type"] not in labels_to_ignore and node_id != expr_id
    ]
    arguments_id = c_ids[0] if c_ids else None

    return build_element_access_node(args, expr_id, arguments_id)
